# -*- coding: utf-8 -*-
"""
Support for AIS local audio

For more details about this component, please refer to the documentation at
https://ai-speaker.com
"""
import asyncio
import logging
import os
import signal
import json
import mimetypes
import subprocess
import time
from homeassistant.components import ais_cloud
from homeassistant.components.ais_dom import ais_global
from .config_flow import configured_drivers

aisCloud = ais_cloud.AisCloudWS()

DOMAIN = "ais_drives_service"
G_LOCAL_FILES_ROOT = "/data/data/pl.sviete.dom/files/home/dom"
G_CLOUD_PREFIX = "dyski-zdalne:"
G_RCLONE_CONF_FILE = "/data/data/pl.sviete.dom/files/home/dom/rclone.conf"
G_RCLONE_CONF = "--config=" + G_RCLONE_CONF_FILE
G_RCLONE_URL_TO_STREAM = "http://127.0.0.1:8080/"
G_DRIVE_CLIENT_ID = None
G_DRIVE_SECRET = None
G_COVER_FILE = "/data/data/pl.sviete.dom/files/home/AIS/www/cover.jpg"
G_RCLONE_REMOTES_LONG = []
_LOGGER = logging.getLogger(__name__)


TYPE_DRIVE = "drive"
TYPE_MEGA = "mega"
DRIVES_TYPES = {
    TYPE_DRIVE: ("Google Drive", "mdi:google-drive"),
    TYPE_MEGA: ("Mega", "mdi:cloud"),
}


def get_pozycji_variety(n):
    if n == 1:
        return str(n) + " pozycja"
    elif n in (2, 3, 4) or (
        n > 20
        and (str(n).endswith("2") or str(n).endswith("3") or str(n).endswith("4"))
    ):
        return str(n) + " pozycje"
    return str(n) + " pozycji"


@asyncio.coroutine
def async_setup(hass, config):
    """Register the service."""
    _LOGGER.info("Initialize the folders and files list.")
    data = hass.data[DOMAIN] = LocalData(hass)
    yield from data.async_load_all()

    # register services
    def browse_path(call):
        data.browse_path(call)

    def refresh_files(call):
        _LOGGER.info("refresh_files")
        data.refresh_files(call)

    def sync_locations(call):
        _LOGGER.info("sync_locations")
        data.sync_locations(call)

    def play_next(call):
        _LOGGER.info("play_next")
        data.play_next(call)

    def play_prev(call):
        _LOGGER.info("play_prev")
        data.play_prev(call)

    def remote_next_item(call):
        _LOGGER.info("remote_next_item")
        data.remote_next_item(True)

    def remote_prev_item(call):
        _LOGGER.info("remote_prev_item")
        data.remote_prev_item(True)

    def remote_select_item(call):
        _LOGGER.info("remote_select_item")
        data.remote_select_item(True)

    def remote_cancel_item(call):
        _LOGGER.info("remote_cancel_item")
        data.remote_cancel_item(True)

    def remote_delete_item(call):
        _LOGGER.info("remote_delete_item")
        data.remote_delete_item(True)

    hass.services.async_register(DOMAIN, "browse_path", browse_path)
    hass.services.async_register(DOMAIN, "refresh_files", refresh_files)
    hass.services.async_register(DOMAIN, "sync_locations", sync_locations)
    hass.services.async_register(DOMAIN, "play_next", play_next)
    hass.services.async_register(DOMAIN, "play_prev", play_prev)
    hass.services.async_register(DOMAIN, "remote_next_item", remote_next_item)
    hass.services.async_register(DOMAIN, "remote_prev_item", remote_prev_item)
    hass.services.async_register(DOMAIN, "remote_select_item", remote_select_item)
    hass.services.async_register(DOMAIN, "remote_cancel_item", remote_cancel_item)
    hass.services.async_register(DOMAIN, "remote_delete_item", remote_delete_item)

    return True


async def async_setup_entry(hass, config_entry):
    """Set up drive as rclone config entry."""
    _LOGGER.info("Set up drive as rclone config entry")
    hass.async_create_task(
        hass.config_entries.async_forward_entry_setup(config_entry, "sensor")
    )
    return True


async def async_unload_entry(hass, config_entry):
    """Unload a config entry."""
    time_now = time.time()
    if config_flow.G_DRIVE_CREATION_TIME_CALL is not None:
        secs = time_now - config_flow.G_DRIVE_CREATION_TIME_CALL
    else:
        secs = 10
    if secs > 5:
        _LOGGER.warning("Remove the drive token from rclone conf: " + str(secs))
        open(G_RCLONE_CONF_FILE, "w").close()
    else:
        _LOGGER.info("Reloading entry: " + str(secs))

    await hass.config_entries.async_forward_entry_unload(config_entry, "sensor")
    return True


def get_remotes_types_by_name(remote_name):
    # return all types
    names = []
    if remote_name is None:
        for obj in DRIVES_TYPES.values():
            names.append(obj[0])
        return names
    # return one type
    drive_type = ""
    for k, item in DRIVES_TYPES.items():
        if item[0] == remote_name:
            drive_type = k
    return drive_type


def rclone_get_remotes_long():
    global G_RCLONE_REMOTES_LONG
    G_RCLONE_REMOTES_LONG = []
    rclone_cmd = ["rclone", "listremotes", "--long", G_RCLONE_CONF]
    proc = subprocess.run(
        rclone_cmd, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    #  will wait for the process to complete and then we are going to return the output
    if "" != proc.stderr:
        _LOGGER.error(
            "Nie można pobrać informacji o połączeniach do dysków: " + proc.stderr
        )
    else:
        for l in proc.stdout.split("\n"):
            if len(l) > 0:
                ri = l.split(":")
                G_RCLONE_REMOTES_LONG.append(
                    {"name": ri[0].strip(), "type": ri[1].strip()}
                )
    return G_RCLONE_REMOTES_LONG


def rclone_get_auth_url(drive_name, drive_type):
    import pexpect

    # code, icon = DRIVES_TYPES[drive_type]
    rclone_cmd = (
        "rclone config create "
        + drive_name
        + " "
        + drive_type
        + " "
        + G_RCLONE_CONF
        + " --drive-client-id="
        + G_DRIVE_CLIENT_ID
        + " --drive-client-secret="
        + G_DRIVE_SECRET
        + " config_is_local false"
    )
    child = pexpect.spawn(rclone_cmd)
    child.expect("Enter verification code>", timeout=10)
    info = child.before
    child.kill(0)
    info = str(info, "utf-8")
    _LOGGER.info(info)
    s = info.find("https://")
    url = info[s:]
    e = url.find("\r")
    url = url[:e]
    return url


def rclone_set_auth_gdrive(drive_name, code):
    try:
        import pexpect

        rclone_cmd = "rclone config " + G_RCLONE_CONF
        child = pexpect.spawn(rclone_cmd)
        # Current remotes:
        child.expect("/q>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # name
        child.expect("name>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(drive_name)
        # Storage
        child.expect("Storage>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("drive")
        # client_id
        child.expect("client_id>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(G_DRIVE_CLIENT_ID)
        # client_secret>
        child.expect("client_secret>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(G_DRIVE_SECRET)
        # scope>
        child.expect("scope>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline()
        # root_folder_id>
        child.expect("root_folder_id>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline()
        # service_account_file>
        child.expect("service_account_file>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline()
        # Edit advanced config? (y/n)
        child.expect("y/n>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # Use auto config? n - Because Remote config
        child.expect("y/n>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # 'Enter verification code>'
        child.expect("Enter verification code>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(code)
        # Configure this as a team drive?
        child.expect("y/n>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # Yes this is OK
        child.expect("y/e/d>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("y")
        # Quit config
        child.expect("e/n/d/r/c/s/q>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("q")
        child.kill(0)
        return "ok"
    except Exception as e:
        return "ERROR: " + str(e)


def rclone_set_auth_mega(drive_name, user, passwd):
    try:
        import pexpect

        rclone_cmd = "rclone config " + G_RCLONE_CONF
        child = pexpect.spawn(rclone_cmd)
        # Current remotes:
        child.expect("/q>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # name
        child.expect("name>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(drive_name)
        # storage
        child.expect("Storage>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("mega")
        # user
        child.expect("user>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(user)
        # yes type in my own password
        child.expect("y/g>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("y")
        # password
        child.expect("password:", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(passwd)
        # confirm password
        child.expect("password:", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline(passwd)
        # Edit advanced config? (y/n)
        child.expect("y/n>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("n")
        # Yes this is OK
        child.expect("y/e/d>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("y")
        # Quit config
        child.expect("e/n/d/r/c/s/q>", timeout=10)
        _LOGGER.info(str(child.before, "utf-8"))
        child.sendline("q")
        #
        child.kill(0)
        return "ok"
    except Exception as e:
        return "ERROR: " + str(e)


def file_tags_extract(path):
    import mutagen.id3
    import mutagen.flac
    import mutagen.mp4

    global G_COVER_FILE
    dir_www = "/data/data/pl.sviete.dom/files/home/AIS/www/"
    dir_name = os.path.basename(os.path.dirname(path)).replace(" ", "")
    ret_path = "/local/" + dir_name + "_cover.jpg"
    f_length = 0
    # file info
    try:
        fi = mutagen.File(path)
        f_length = str(fi.info.length)
    except Exception as e:
        _LOGGER.error("Error " + str(e))

    if G_COVER_FILE == dir_www + dir_name + "_cover.jpg":
        return ret_path, f_length
    # remove all .jpg
    jpgs = os.listdir(dir_www)
    for jpg in jpgs:
        if jpg.endswith("_cover.jpg"):
            os.remove(os.path.join(dir_www, jpg))

    # generate the cover
    G_COVER_FILE = dir_www + dir_name + "_cover.jpg"
    try:
        id3 = mutagen.id3.ID3(path)
        open(G_COVER_FILE, "wb").write(id3.getall("APIC")[0].data)
    except mutagen.id3.ID3NoHeaderError:
        try:
            flac = mutagen.flac.FLAC(path)
            open(G_COVER_FILE, "wb").write(flac.pictures[0].data)
        except mutagen.flac.FLACNoHeaderError:
            try:
                mp4 = mutagen.mp4.MP4(path)
                open(G_COVER_FILE, "wb").write(mp4["covr"][0])
            except Exception as e:
                _LOGGER.error("Error " + str(e))

    return ret_path, f_length


class LocalData:
    """Class to hold local folders and files data."""

    def __init__(self, hass):
        """Initialize the books authors."""
        self.hass = hass
        self.selected_item_idx = 0
        self.folders_json = []
        self.current_path = os.path.abspath(G_LOCAL_FILES_ROOT)
        self.rclone_url_to_stream = None
        self.rclone_pexpect_stream = None
        self.file_path = None
        self.seek_position = 0

    def beep(self):
        self.hass.services.call(
            "ais_ai_service", "publish_command_to_frame", {"key": "tone", "val": 97}
        )

    def say(self, text):
        self.hass.services.call("ais_ai_service", "say_it", {"text": text})

    def refresh_files(self, call):
        pass

    def play_file(self, say):
        mime_type = mimetypes.MimeTypes().guess_type(self.current_path)[0]
        if mime_type is None:
            mime_type = ""
        if mime_type.startswith("text/"):
            self.say("Czytam: ")
            with open(self.current_path) as file:
                self.say(file.read())
        elif mime_type.startswith("audio/"):
            _url = self.current_path
            # TODO search the album and title ...
            album_cover_path, file_length = file_tags_extract(self.current_path)
            _audio_info = {
                "NAME": os.path.basename(self.current_path),
                "MEDIA_SOURCE": ais_global.G_AN_LOCAL,
                "ALBUM_NAME": os.path.basename(os.path.dirname(self.current_path)),
                "IMAGE_URL": album_cover_path,
                "DURATION": file_length,
                "media_content_id": _url,
                "lookup_url": self.current_path,
                "media_position_ms": self.seek_position,
            }
            _audio_info = json.dumps(_audio_info)

            if _url is not None:
                # set url stream image and title
                self.hass.services.call(
                    "media_player",
                    "play_media",
                    {
                        "entity_id": "media_player.wbudowany_glosnik",
                        "media_content_type": "ais_content_info",
                        "media_content_id": _audio_info,
                    },
                )
                # seek position
                self.seek_position = 0
        else:
            _LOGGER.info(
                "Tego typu plików jeszcze nie obsługuję." + str(self.current_path)
            )
            self.say("Tego typu plików jeszcze nie obsługuję.")

        self.dispalay_current_path()

    def display_root_items(self, say):
        self.hass.states.set(
            "sensor.ais_drives",
            "",
            {
                "files": [
                    {
                        "name": "Dysk wewnętrzny",
                        "icon": "harddisk",
                        "path": G_LOCAL_FILES_ROOT + "/dysk-wewnętrzny",
                    },
                    {
                        "name": "Dyski zewnętrzne",
                        "icon": "sd",
                        "path": G_LOCAL_FILES_ROOT + "/dyski-zewnętrzne",
                    },
                    {
                        "name": "Dyski zdalne",
                        "icon": "onedrive",
                        "path": G_CLOUD_PREFIX,
                    },
                ]
            },
        )
        if say:
            self.say("Dysk wewnętrzny")

    def dispalay_current_path(self):
        state = self.hass.states.get("sensor.ais_drives")
        items_info = state.attributes
        self.hass.states.set(
            "sensor.ais_drives",
            self.current_path.replace(G_LOCAL_FILES_ROOT, ""),
            items_info,
        )

    # browse files on local folder
    def display_current_items(self, say):
        local_items = []
        try:
            local_items = os.scandir(self.current_path)
        except Exception as e:
            _LOGGER.error("list_dir error: " + str(e))
        si = sorted(local_items, key=lambda en: en.name)
        items_info = [
            {"name": ".", "icon": "", "path": G_LOCAL_FILES_ROOT},
            {"name": "..", "icon": "", "path": ".."},
        ]
        for i in si:
            items_info.append(
                {"name": i.name, "icon": self.get_icon(i), "path": i.path}
            )
        self.hass.states.set(
            "sensor.ais_drives",
            self.current_path.replace(G_LOCAL_FILES_ROOT, ""),
            {"files": items_info},
        )
        if say:
            slen = len(si)
            self.say(get_pozycji_variety(slen))

        # call from bookmarks now (since we have files from folder) we need to play the file
        if self.file_path is not None:
            self.hass.services.call(
                "ais_drives_service",
                "browse_path",
                {"path": self.file_path, "seek_position": self.seek_position},
            )

    def display_current_remotes(self, remotes):
        items_info = [
            {"name": ".", "icon": "", "path": G_LOCAL_FILES_ROOT},
            {"name": "..", "icon": "", "path": ".."},
        ]
        for i in remotes:
            if i["type"] == "drive":
                icon = "folder-google-drive"
            else:
                icon = "onedrive"
            items_info.append(
                {
                    "name": i["name"],
                    "icon": icon,
                    "path": self.current_path + i["name"] + ":",
                }
            )
        self.hass.states.set(
            "sensor.ais_drives", self.current_path, {"files": items_info}
        )

    # browse files on cloud folder
    def display_current_remote_items(self, say):
        items_info = [
            {"name": ".", "icon": "", "path": G_LOCAL_FILES_ROOT},
            {"name": "..", "icon": "", "path": ".."},
        ]

        if self.current_path.endswith(":"):
            for remote in G_RCLONE_REMOTES_LONG:
                if "dyski-zdalne:" + remote["name"] + ":" == self.current_path:
                    if remote["type"] == "drive":
                        items_info.append(
                            {
                                "name": ais_global.G_DRIVE_SHARED_WITH_ME,
                                "icon": "account-supervisor-circle",
                                "path": self.current_path
                                + ais_global.G_DRIVE_SHARED_WITH_ME,
                            }
                        )
                        break
        for item in self.folders_json:
            if self.current_path.endswith(":"):
                path = self.current_path + item["Path"]
            else:
                path = self.current_path + "/" + item["Path"]

            l_icon = "file-outline"
            if item["IsDir"]:
                l_icon = "folder-google-drive"
            else:
                if "MimeType" in item:
                    if item["MimeType"].startswith("text/"):
                        l_icon = "file-document-outline"
                    elif item["MimeType"].startswith("audio/"):
                        l_icon = "music-circle"
                    elif item["MimeType"].startswith("video/"):
                        l_icon = "file-video-outline"
            items_info.append({"name": item["Path"][:50], "icon": l_icon, "path": path})
        self.hass.states.set(
            "sensor.ais_drives", self.current_path, {"files": items_info}
        )
        if say:
            jlen = len(self.folders_json)
            self.say(get_pozycji_variety(jlen))

        # call from bookmarks now (since we have files from folder) we need to play the file
        if self.file_path is not None:
            self.hass.services.call(
                "ais_drives_service",
                "browse_path",
                {"path": self.file_path, "seek_position": self.seek_position},
            )

    def get_icon(self, entry):
        if entry.is_dir():
            return "folder"
        elif entry.name.lower().endswith(".txt"):
            return "file-document-outline"
        elif entry.name.lower().endswith((".mp3", ".wav", ".mp4", ".flv")):
            return "music-circle"

    def browse_path(self, call):
        """Load subfolders for the selected folder."""
        if "path" not in call.data:
            _LOGGER.error("No path")
            return
        self.file_path = None
        self.seek_position = 0
        say = True
        if "file_path" in call.data:
            self.file_path = call.data["file_path"]
            say = False
        if "seek_position" in call.data:
            self.seek_position = call.data["seek_position"]
            say = False
        self._browse_path(call.data["path"], say)

    def _browse_path(self, path, say):
        if len(path.strip()) == 0:
            self.say("Wybierz pozycję do przeglądania")
        if path == "..":
            # check if this is cloud drive
            if self.is_rclone_path(self.current_path):
                if self.current_path == G_CLOUD_PREFIX:
                    self.current_path = G_LOCAL_FILES_ROOT
                elif self.current_path == G_CLOUD_PREFIX + self.rclone_remote_from_path(
                    self.current_path
                ):
                    self.current_path = G_CLOUD_PREFIX
                elif self.current_path.count("/") == 0:
                    k = self.current_path.rfind(":")
                    self.current_path = self.current_path[: k + 1]
                else:
                    if self.rclone_is_dir(self.current_path):
                        k = self.current_path.rfind("/")
                        self.current_path = self.current_path[:k]
                    else:
                        k = self.current_path.rfind("/")
                        self.current_path = self.current_path[:k]
                        if self.current_path.count("/") > 0:
                            k = self.current_path.rfind("/")
                            self.current_path = self.current_path[:k]
                        else:
                            k = self.current_path.rfind(":")
                            self.current_path = self.current_path[: k + 1]

            # local drive
            else:
                if os.path.isfile(self.current_path):
                    k = self.current_path.rfind(
                        "/" + os.path.basename(self.current_path)
                    )
                    self.current_path = self.current_path[:k]
                k = self.current_path.rfind("/" + os.path.basename(self.current_path))
                self.current_path = self.current_path[:k]
        else:
            self.current_path = path

        if self.current_path.startswith(G_CLOUD_PREFIX):
            self.rclone_browse(self.current_path, say)
            self.selected_item_idx = 0
            return

        if self.current_path == G_LOCAL_FILES_ROOT:
            self.display_root_items(say)
            self.selected_item_idx = 0
            return

        if os.path.isdir(self.current_path):
            self.display_current_items(say)
            self.selected_item_idx = 0
            return
        else:
            # file was selected, check mimetype and play if possible
            self.play_file(say)

    def is_rclone_path(self, path):
        if path.startswith(G_CLOUD_PREFIX):
            return True
        return False

    def rclone_remote_from_path(self, path):
        remote = path.replace(G_CLOUD_PREFIX, "")
        k = remote.find(":")
        remote = remote[: k + 1]
        return remote

    def rclone_fix_permissions(self):
        command = 'su -c "chmod -R 777 /sdcard/rclone"'
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        # process.wait()

    def rclone_append_listremotes(self):
        remotes = rclone_get_remotes_long()
        self.display_current_remotes(remotes)
        if len(remotes) == 0:
            self.say(
                "Nie masz żadnych dysków zdalnych. "
                "Dodaj połączenie do dysku zdalnego za pomocą konfiguratora w aplikacji."
            )
        else:
            self.say(
                "Mamy "
                + get_pozycji_variety(len(remotes))
                + " Wybierz dysk który mam przeglądać."
            )

    def rclone_browse_folder(self, path, say):
        if ais_global.G_DRIVE_SHARED_WITH_ME in path:
            if ais_global.G_DRIVE_SHARED_WITH_ME + "/" in path:
                path = path.replace(ais_global.G_DRIVE_SHARED_WITH_ME + "/", "")
            else:
                path = path.replace(ais_global.G_DRIVE_SHARED_WITH_ME, "")
            rclone_cmd = [
                "rclone",
                "lsjson",
                path,
                G_RCLONE_CONF,
                "--drive-formats=txt",
                "--drive-shared-with-me",
            ]
        else:
            rclone_cmd = [
                "rclone",
                "lsjson",
                path,
                G_RCLONE_CONF,
                "--drive-formats=txt",
            ]
        proc = subprocess.run(
            rclone_cmd, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        #  will wait for the process to complete and then we are going to return the output
        if "" != proc.stderr:
            self.say("Nie można pobrać zawartości folderu " + path + " " + proc.stderr)
        else:
            self.folders_json = json.loads(proc.stdout)
            self.display_current_remote_items(say)

    def rclone_copy_and_read(self, path, item_path):
        # clear .temp files
        files = os.listdir(G_LOCAL_FILES_ROOT + "/.temp/")
        for file in files:
            os.remove(os.path.join(G_LOCAL_FILES_ROOT + "/.temp/", file))

        if ais_global.G_DRIVE_SHARED_WITH_ME in path:
            rclone_cmd = [
                "rclone",
                "copy",
                path.replace(ais_global.G_DRIVE_SHARED_WITH_ME, ""),
                G_LOCAL_FILES_ROOT + "/.temp/",
                G_RCLONE_CONF,
                "--drive-formats=txt",
                "--drive-shared-with-me",
            ]
        else:
            rclone_cmd = [
                "rclone",
                "copy",
                path,
                G_LOCAL_FILES_ROOT + "/.temp/",
                G_RCLONE_CONF,
                "--drive-formats=txt",
            ]
        proc = subprocess.run(
            rclone_cmd, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        if "" != proc.stderr:
            self.say("Nie udało się pobrać pliku " + proc.stderr)
        else:
            try:
                with open(G_LOCAL_FILES_ROOT + "/.temp/" + item_path) as file:
                    self.say(file.read())
            except Exception as e:
                self.say("Nie udało się otworzyć pliku ")

    def rclone_play_the_stream(self):
        _audio_info = {
            "NAME": os.path.basename(self.rclone_url_to_stream),
            "MEDIA_SOURCE": ais_global.G_AN_LOCAL,
            "ALBUM_NAME": os.path.basename(os.path.dirname(self.current_path)),
            "media_content_id": self.rclone_url_to_stream,
            "lookup_url": self.current_path,
            "media_position_ms": self.seek_position,
        }
        _audio_info = json.dumps(_audio_info)
        # to set the stream image and title
        self.hass.services.call(
            "media_player",
            "play_media",
            {
                "entity_id": "media_player.wbudowany_glosnik",
                "media_content_type": "ais_content_info",
                "media_content_id": _audio_info,
            },
        )
        # seek position
        self.seek_position = 0

    def check_kill_process(self, pstring):
        for line in os.popen("ps ax | grep " + pstring + " | grep -v grep"):
            fields = line.split()
            pid = fields[0]
            os.kill(int(pid), signal.SIGKILL)

    def rclone_serve_and_play_the_stream(self, path, item_path):
        # serve and play
        if ais_global.G_DRIVE_SHARED_WITH_ME in path:
            path = path.replace(ais_global.G_DRIVE_SHARED_WITH_ME, "")
        rclone_cmd = (
            "rclone serve http '" + path + "' " + G_RCLONE_CONF + " --addr=:8080"
        )

        self.rclone_url_to_stream = G_RCLONE_URL_TO_STREAM + str(item_path)
        import pexpect

        try:
            if self.rclone_pexpect_stream is not None:
                self.rclone_pexpect_stream.kill(0)
                self.check_kill_process("rclone")
            self.rclone_pexpect_stream = pexpect.spawn(rclone_cmd)
            # Current remotes:
            self.rclone_pexpect_stream.expect(
                ["Serving on", "Failed to", pexpect.EOF], timeout=10
            )
            _LOGGER.info(str(self.rclone_pexpect_stream.before, "utf-8"))
            if self.rclone_pexpect_stream == 0:
                _LOGGER.info("Serving stream")
            elif self.rclone_pexpect_stream == 1:
                _LOGGER.info("Problem, kill rclone")
                self.rclone_pexpect_stream.kill(0)
                self.check_kill_process("rclone")
            elif self.rclone_pexpect_stream == 2:
                _LOGGER.info("EOF")
            self.rclone_play_the_stream()
        except Exception as e:
            _LOGGER.info("Rclone: " + str(e))

    def rclone_is_dir(self, path):
        # check if path is dir or file
        for item in self.folders_json:
            if path.endswith(item["Path"]):
                return item["IsDir"]
        return True

    def rclone_browse(self, path, say):
        if path == G_CLOUD_PREFIX:
            self.rclone_append_listremotes()
            return
        is_dir = None
        mime_type = ""
        item_name = ""
        item_path = ""
        # check what was selected file or folder
        for item in self.folders_json:
            # now check if item is a dictionary
            if path.endswith(item["Path"]):
                item_path = item["Path"]
                is_dir = item["IsDir"]
                item_name = item["Name"]
                if "MimeType" in item:
                    mime_type = item["MimeType"]
                break
        if is_dir is None:
            is_dir = True
        if is_dir:
            # browse the cloud drive
            path = path.replace(G_CLOUD_PREFIX, "", 1)
            self.say("Pobieram")
            self.rclone_browse_folder(path, say)
        else:
            self.dispalay_current_path()
            # file was selected, check the MimeType
            # "MimeType":"audio/mp3" and "text/plain" are supported
            path = path.replace(G_CLOUD_PREFIX, "")
            if mime_type is None:
                mime_type = ""
            if (
                mime_type.startswith("audio/")
                or mime_type.startswith("video/")
                or mime_type.startswith("application/")
            ):
                # StreamTask().execute(fileItem);
                self.say("Pobieram i odtwarzam: " + str(item_name))
                self.rclone_serve_and_play_the_stream(path, item_path)
            elif mime_type.startswith("text/"):
                self.say("Pobieram i czytam: " + str(item_name))
                self.rclone_copy_and_read(path, item_path)
            else:
                self.say("Jeszcze nie obsługuję plików typu: " + str(mime_type))

    def sync_locations(self, call):
        if "source_path" not in call.data:
            _LOGGER.error("No source_path")
            return []
        if "dest_path" not in call.data:
            _LOGGER.error("No dest_path")
            return []
        if "say" in call.data:
            say = call.data["say"]
        else:
            say = False

        if say:
            self.say(
                "Synchronizuję lokalizację "
                + call.data["source_path"]
                + " z "
                + call.data["dest_path"]
                + " modyfikuję tylko "
                + call.data["source_path"]
            )

        rclone_cmd = [
            "rclone",
            "sync",
            call.data["source_path"],
            call.data["dest_path"],
            "--transfers=1",
            "--stats=0",
            G_RCLONE_CONF,
        ]
        proc = subprocess.run(
            rclone_cmd, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        #  will wait for the process to complete and then we are going to return the output
        if "" != proc.stderr:
            self.say("Błąd podczas synchronizacji: " + proc.stderr)
        else:
            self.say("Synchronizacja zakończona.")

    def play_next(self, call):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        l_idx = 0
        i = 0
        for f in files:
            i = i + 1
            if f["path"].replace(G_LOCAL_FILES_ROOT, "") == state.state:
                l_idx = i
        if l_idx == len(files):
            l_idx = min(2, len(files))
        self._browse_path(files[l_idx]["path"], True)

    def play_prev(self, call):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        l_idx = 0
        i = 0
        for f in files:
            i = i + 1
            if f["path"].replace(G_LOCAL_FILES_ROOT, "") == state.state:
                l_idx = i
        if l_idx == 3:
            l_idx = len(files) - 1
        else:
            l_idx = l_idx - 2
        self._browse_path(files[l_idx]["path"], True)

    def get_item_name(self, path):
        path = path.rstrip(":")
        if path.count("/") > 0:
            name = path.split("/").pop()
        else:
            name = path.split(":").pop()
        name = name.replace(":", "")
        name = name.replace("-", " ")
        return name

    def remote_next_item(self, say):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        if len(state.state) == 0:
            if self.selected_item_idx == len(files) - 1:
                self.selected_item_idx = 0
            else:
                self.selected_item_idx = self.selected_item_idx + 1
        else:
            if len(files) == 2:
                self.say("brak pozycji")
                return

            if self.selected_item_idx == len(files) - 1:
                self.selected_item_idx = 2
            else:
                self.selected_item_idx = max(self.selected_item_idx + 1, 2)

        if say:
            name = self.get_item_name(files[self.selected_item_idx]["path"])
            self.say(name)

    def remote_prev_item(self, say):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        if len(state.state) == 0:
            if self.selected_item_idx < 1:
                self.selected_item_idx = len(files) - 1
            else:
                self.selected_item_idx = self.selected_item_idx - 1
        else:
            if len(files) == 2:
                self.say("brak pozycji")
                return

            if self.selected_item_idx < 3:
                self.selected_item_idx = len(files) - 1
            else:
                self.selected_item_idx = self.selected_item_idx - 1
        if say:
            name = self.get_item_name(files[self.selected_item_idx]["path"])
            self.say(name)

    def remote_select_item(self, say):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        if state.state is None or self.selected_item_idx is None:
            self.selected_item_idx = 0
        if files[self.selected_item_idx]["path"] == G_LOCAL_FILES_ROOT:
            self.say("Wybierz pozycję")
        else:
            self._browse_path(files[self.selected_item_idx]["path"], say)

    def remote_cancel_item(self, say):
        self._browse_path("..", say)

    def remote_delete_item(self, say):
        state = self.hass.states.get("sensor.ais_drives")
        attr = state.attributes
        files = attr.get("files", [])
        if state.state is None or self.selected_item_idx is None:
            self.say("Brak pozycji do usunięcia")
        if files[self.selected_item_idx]["path"] == G_LOCAL_FILES_ROOT:
            self.say("Tej pozycji nie można usunąć")
        else:
            # delete file or folder
            path = files[self.selected_item_idx]["path"]
            if path.startswith(G_LOCAL_FILES_ROOT) and path != G_LOCAL_FILES_ROOT:
                if os.path.isdir(path):
                    # remove dir
                    import shutil

                    shutil.rmtree(path)
                    self.say("Usuwam folder " + files[self.selected_item_idx]["name"])
                else:
                    # remove file
                    os.remove(path)
                    self.say("Usuwam plik " + files[self.selected_item_idx]["name"])
                # browse this patch again
                from pathlib import Path

                parent_path = str(Path(path).parent)
                self._browse_path(parent_path, False)

            else:
                self.say("Tej pozycji nie można usunąć")

    @asyncio.coroutine
    def async_load_all(self):
        """Load all the folders and files."""

        def load():
            self.display_root_items(False)
            global G_DRIVE_SECRET, G_DRIVE_CLIENT_ID
            try:
                ws_resp = aisCloud.key("gdrive_client_id")
                json_ws_resp = ws_resp.json()
                G_DRIVE_CLIENT_ID = json_ws_resp["key"]
                ws_resp = aisCloud.key("gdrive_secret")
                json_ws_resp = ws_resp.json()
                G_DRIVE_SECRET = json_ws_resp["key"]
            except Exception as e:
                _LOGGER.error("Error " + str(e))
                ais_global.G_OFFLINE_MODE = True

        yield from self.hass.async_add_job(load)



"""This component provides HA sensor support for Travis CI framework."""
import logging
from datetime import timedelta

import voluptuous as vol

import homeassistant.helpers.config_validation as cv
from homeassistant.components.sensor import PLATFORM_SCHEMA
from homeassistant.const import (
    ATTR_ATTRIBUTION,
    CONF_API_KEY,
    CONF_SCAN_INTERVAL,
    CONF_MONITORED_CONDITIONS,
)
from homeassistant.helpers.entity import Entity

_LOGGER = logging.getLogger(__name__)

ATTRIBUTION = "Information provided by https://travis-ci.org/"

CONF_BRANCH = "branch"
CONF_REPOSITORY = "repository"

DEFAULT_BRANCH_NAME = "master"

SCAN_INTERVAL = timedelta(seconds=30)

# sensor_type [ description, unit, icon ]
SENSOR_TYPES = {
    "last_build_id": ["Last Build ID", "", "mdi:account-card-details"],
    "last_build_duration": ["Last Build Duration", "sec", "mdi:timelapse"],
    "last_build_finished_at": ["Last Build Finished At", "", "mdi:timetable"],
    "last_build_started_at": ["Last Build Started At", "", "mdi:timetable"],
    "last_build_state": ["Last Build State", "", "mdi:github-circle"],
    "state": ["State", "", "mdi:github-circle"],
}

NOTIFICATION_ID = "travisci"
NOTIFICATION_TITLE = "Travis CI Sensor Setup"

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_API_KEY): cv.string,
        vol.Required(CONF_MONITORED_CONDITIONS, default=list(SENSOR_TYPES)): vol.All(
            cv.ensure_list, [vol.In(SENSOR_TYPES)]
        ),
        vol.Required(CONF_BRANCH, default=DEFAULT_BRANCH_NAME): cv.string,
        vol.Optional(CONF_REPOSITORY, default=[]): vol.All(cv.ensure_list, [cv.string]),
        vol.Optional(CONF_SCAN_INTERVAL, default=SCAN_INTERVAL): cv.time_period,
    }
)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the Travis CI sensor."""
    from travispy import TravisPy
    from travispy.errors import TravisError

    token = config.get(CONF_API_KEY)
    repositories = config.get(CONF_REPOSITORY)
    branch = config.get(CONF_BRANCH)

    try:
        travis = TravisPy.github_auth(token)
        user = travis.user()

    except TravisError as ex:
        _LOGGER.error("Unable to connect to Travis CI service: %s", str(ex))
        hass.components.persistent_notification.create(
            "Error: {}<br />"
            "You will need to restart hass after fixing."
            "".format(ex),
            title=NOTIFICATION_TITLE,
            notification_id=NOTIFICATION_ID,
        )
        return False

    sensors = []

    # non specific repository selected, then show all associated
    if not repositories:
        all_repos = travis.repos(member=user.login)
        repositories = [repo.slug for repo in all_repos]

    for repo in repositories:
        if "/" not in repo:
            repo = "{0}/{1}".format(user.login, repo)

        for sensor_type in config.get(CONF_MONITORED_CONDITIONS):
            sensors.append(TravisCISensor(travis, repo, user, branch, sensor_type))

    add_entities(sensors, True)
    return True


class TravisCISensor(Entity):
    """Representation of a Travis CI sensor."""

    def __init__(self, data, repo_name, user, branch, sensor_type):
        """Initialize the sensor."""
        self._build = None
        self._sensor_type = sensor_type
        self._data = data
        self._repo_name = repo_name
        self._user = user
        self._branch = branch
        self._state = None
        self._name = "{0} {1}".format(
            self._repo_name, SENSOR_TYPES[self._sensor_type][0]
        )

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement of this entity, if any."""
        return SENSOR_TYPES[self._sensor_type][1]

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def device_state_attributes(self):
        """Return the state attributes."""
        attrs = {}
        attrs[ATTR_ATTRIBUTION] = ATTRIBUTION

        if self._build and self._state is not None:
            if self._user and self._sensor_type == "state":
                attrs["Owner Name"] = self._user.name
                attrs["Owner Email"] = self._user.email
            else:
                attrs["Committer Name"] = self._build.commit.committer_name
                attrs["Committer Email"] = self._build.commit.committer_email
                attrs["Commit Branch"] = self._build.commit.branch
                attrs["Committed Date"] = self._build.commit.committed_at
                attrs["Commit SHA"] = self._build.commit.sha

        return attrs

    @property
    def icon(self):
        """Return the icon to use in the frontend, if any."""
        return SENSOR_TYPES[self._sensor_type][2]

    def update(self):
        """Get the latest data and updates the states."""
        _LOGGER.debug("Updating sensor %s", self._name)

        repo = self._data.repo(self._repo_name)
        self._build = self._data.build(repo.last_build_id)

        if self._build:
            if self._sensor_type == "state":
                branch_stats = self._data.branch(self._branch, self._repo_name)
                self._state = branch_stats.state

            else:
                param = self._sensor_type.replace("last_build_", "")
                self._state = getattr(self._build, param)



"""Icon helper methods."""
from typing import Optional


def icon_for_battery_level(
    battery_level: Optional[int] = None, charging: bool = False
) -> str:
    """Return a battery icon valid identifier."""
    icon = "mdi:battery"
    if battery_level is None:
        return icon + "-unknown"
    if charging and battery_level > 10:
        icon += "-charging-{}".format(int(round(battery_level / 20 - 0.01)) * 20)
    elif charging:
        icon += "-outline"
    elif battery_level <= 5:
        icon += "-alert"
    elif 5 < battery_level < 95:
        icon += "-{}".format(int(round(battery_level / 10 - 0.01)) * 10)
    return icon



"""Classes to help gather user submissions."""
import logging
from typing import (
    Dict,
    Any,
    Callable,
    Hashable,
    List,
    Optional,
)  # noqa pylint: disable=unused-import
import uuid
import voluptuous as vol
from .core import callback, HomeAssistant
from .exceptions import HomeAssistantError

_LOGGER = logging.getLogger(__name__)

RESULT_TYPE_FORM = "form"
RESULT_TYPE_CREATE_ENTRY = "create_entry"
RESULT_TYPE_ABORT = "abort"
RESULT_TYPE_EXTERNAL_STEP = "external"
RESULT_TYPE_EXTERNAL_STEP_DONE = "external_done"

# Event that is fired when a flow is progressed via external source.
EVENT_DATA_ENTRY_FLOW_PROGRESSED = "data_entry_flow_progressed"


class FlowError(HomeAssistantError):
    """Error while configuring an account."""


class UnknownHandler(FlowError):
    """Unknown handler specified."""


class UnknownFlow(FlowError):
    """Uknown flow specified."""


class UnknownStep(FlowError):
    """Unknown step specified."""


class FlowManager:
    """Manage all the flows that are in progress."""

    def __init__(
        self,
        hass: HomeAssistant,
        async_create_flow: Callable,
        async_finish_flow: Callable,
    ) -> None:
        """Initialize the flow manager."""
        self.hass = hass
        self._progress = {}  # type: Dict[str, Any]
        self._async_create_flow = async_create_flow
        self._async_finish_flow = async_finish_flow

    @callback
    def async_progress(self) -> List[Dict]:
        """Return the flows in progress."""
        return [
            {"flow_id": flow.flow_id, "handler": flow.handler, "context": flow.context}
            for flow in self._progress.values()
        ]

    async def async_init(
        self, handler: Hashable, *, context: Optional[Dict] = None, data: Any = None
    ) -> Any:
        """Start a configuration flow."""
        if context is None:
            context = {}
        flow = await self._async_create_flow(handler, context=context, data=data)
        flow.hass = self.hass
        flow.handler = handler
        flow.flow_id = uuid.uuid4().hex
        flow.context = context
        self._progress[flow.flow_id] = flow

        return await self._async_handle_step(flow, flow.init_step, data)

    async def async_configure(
        self, flow_id: str, user_input: Optional[Dict] = None
    ) -> Any:
        """Continue a configuration flow."""
        flow = self._progress.get(flow_id)

        if flow is None:
            raise UnknownFlow

        cur_step = flow.cur_step

        if cur_step.get("data_schema") is not None and user_input is not None:
            user_input = cur_step["data_schema"](user_input)

        result = await self._async_handle_step(flow, cur_step["step_id"], user_input)

        if cur_step["type"] == RESULT_TYPE_EXTERNAL_STEP:
            if result["type"] not in (
                RESULT_TYPE_EXTERNAL_STEP,
                RESULT_TYPE_EXTERNAL_STEP_DONE,
            ):
                raise ValueError(
                    "External step can only transition to "
                    "external step or external step done."
                )

            # If the result has changed from last result, fire event to update
            # the frontend.
            if cur_step["step_id"] != result.get("step_id"):
                # Tell frontend to reload the flow state.
                self.hass.bus.async_fire(
                    EVENT_DATA_ENTRY_FLOW_PROGRESSED,
                    {"handler": flow.handler, "flow_id": flow_id, "refresh": True},
                )

        return result

    @callback
    def async_abort(self, flow_id: str) -> None:
        """Abort a flow."""
        if self._progress.pop(flow_id, None) is None:
            raise UnknownFlow

    async def _async_handle_step(
        self, flow: Any, step_id: str, user_input: Optional[Dict]
    ) -> Dict:
        """Handle a step of a flow."""
        method = "async_step_{}".format(step_id)

        if not hasattr(flow, method):
            self._progress.pop(flow.flow_id)
            raise UnknownStep(
                "Handler {} doesn't support step {}".format(
                    flow.__class__.__name__, step_id
                )
            )

        result = await getattr(flow, method)(user_input)  # type: Dict

        if result["type"] not in (
            RESULT_TYPE_FORM,
            RESULT_TYPE_EXTERNAL_STEP,
            RESULT_TYPE_CREATE_ENTRY,
            RESULT_TYPE_ABORT,
            RESULT_TYPE_EXTERNAL_STEP_DONE,
        ):
            raise ValueError(
                "Handler returned incorrect type: {}".format(result["type"])
            )

        if result["type"] in (
            RESULT_TYPE_FORM,
            RESULT_TYPE_EXTERNAL_STEP,
            RESULT_TYPE_EXTERNAL_STEP_DONE,
        ):
            flow.cur_step = result
            return result

        # We pass a copy of the result because we're mutating our version
        result = await self._async_finish_flow(flow, dict(result))

        # _async_finish_flow may change result type, check it again
        if result["type"] == RESULT_TYPE_FORM:
            flow.cur_step = result
            return result

        # Abort and Success results both finish the flow
        self._progress.pop(flow.flow_id)

        return result


class FlowHandler:
    """Handle the configuration flow of a component."""

    # Set by flow manager
    flow_id: Optional[str] = None
    hass: Optional[HomeAssistant] = None
    handler = None
    cur_step: Optional[Dict[str, str]] = None
    context: Dict

    # Set by _async_create_flow callback
    init_step = "init"

    # Set by developer
    VERSION = 1

    @callback
    def async_show_form(
        self,
        *,
        step_id: str,
        data_schema: vol.Schema = None,
        errors: Optional[Dict] = None,
        description_placeholders: Optional[Dict] = None,
    ) -> Dict:
        """Return the definition of a form to gather user input."""
        return {
            "type": RESULT_TYPE_FORM,
            "flow_id": self.flow_id,
            "handler": self.handler,
            "step_id": step_id,
            "data_schema": data_schema,
            "errors": errors,
            "description_placeholders": description_placeholders,
        }

    @callback
    def async_create_entry(
        self,
        *,
        title: str,
        data: Dict,
        description: Optional[str] = None,
        description_placeholders: Optional[Dict] = None,
    ) -> Dict:
        """Finish config flow and create a config entry."""
        return {
            "version": self.VERSION,
            "type": RESULT_TYPE_CREATE_ENTRY,
            "flow_id": self.flow_id,
            "handler": self.handler,
            "title": title,
            "data": data,
            "description": description,
            "description_placeholders": description_placeholders,
        }

    @callback
    def async_abort(
        self, *, reason: str, description_placeholders: Optional[Dict] = None
    ) -> Dict:
        """Abort the config flow."""
        return {
            "type": RESULT_TYPE_ABORT,
            "flow_id": self.flow_id,
            "handler": self.handler,
            "reason": reason,
            "description_placeholders": description_placeholders,
        }

    @callback
    def async_external_step(
        self, *, step_id: str, url: str, description_placeholders: Optional[Dict] = None
    ) -> Dict:
        """Return the definition of an external step for the user to take."""
        return {
            "type": RESULT_TYPE_EXTERNAL_STEP,
            "flow_id": self.flow_id,
            "handler": self.handler,
            "step_id": step_id,
            "url": url,
            "description_placeholders": description_placeholders,
        }

    @callback
    def async_external_step_done(self, *, next_step_id: str) -> Dict:
        """Return the definition of an external step for the user to take."""
        return {
            "type": RESULT_TYPE_EXTERNAL_STEP_DONE,
            "flow_id": self.flow_id,
            "handler": self.handler,
            "step_id": next_step_id,
        }



"""
Support for interacting with Spotify.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/media_player.spotify/
"""
import logging
import asyncio
from homeassistant.components.ais_dom import ais_global
from homeassistant.components import ais_cloud
from .config_flow import configured_service, setUrl
from homeassistant.util.async_ import run_coroutine_threadsafe

aisCloud = ais_cloud.AisCloudWS()

_LOGGER = logging.getLogger(__name__)

AUTH_CALLBACK_NAME = "api:ais_spotify_service"
AUTH_CALLBACK_PATH = ""

CONF_ALIASES = "aliases"
CONF_CACHE_PATH = "cache_path"
CONF_CLIENT_ID = "client_id"
CONF_CLIENT_SECRET = "client_secret"
AIS_SPOTIFY_TOKEN = None
OAUTH_CLIENT_ID = None

CONFIGURATOR_DESCRIPTION = "Aby połączyć swoje konto Spotify, kliknij link:"
CONFIGURATOR_LINK_NAME = "Połącz konto Spotify"
CONFIGURATOR_SUBMIT_CAPTION = "OK, dostęp dodany!"

DEFAULT_CACHE_PATH = ".dom/.ais-dom-spotify-token-cache"
DEFAULT_NAME = "Spotify"
DEPENDENCIES = ["http"]
DOMAIN = "ais_spotify_service"

ICON = "mdi:spotify"

SCOPE = "app-remote-control streaming user-read-email"
_CONFIGURING = {}


@asyncio.coroutine
def async_setup(hass, config):
    """Set up the Spotify platform."""
    import spotipy.oauth2
    import json

    global AIS_SPOTIFY_TOKEN

    # info about discovery
    async def do_the_spotify_disco(service):
        """ Called when a Spotify integration has been discovered. """
        await hass.config_entries.flow.async_init(
            "ais_spotify_service", context={"source": "discovery"}, data={}
        )
        await hass.async_block_till_done()

    try:
        ws_resp = aisCloud.key("spotify_oauth")
        json_ws_resp = ws_resp.json()
        spotify_redirect_url = json_ws_resp["SPOTIFY_REDIRECT_URL"]
        spotify_client_id = json_ws_resp["SPOTIFY_CLIENT_ID"]
        spotify_client_secret = json_ws_resp["SPOTIFY_CLIENT_SECRET"]
        spotify_scope = json_ws_resp["SPOTIFY_SCOPE"]
        try:
            ws_resp = aisCloud.key("spotify_token")
            key = ws_resp.json()["key"]
            AIS_SPOTIFY_TOKEN = json.loads(key)
        except:
            AIS_SPOTIFY_TOKEN = None
            _LOGGER.info("No AIS_SPOTIFY_TOKEN")
    except Exception as e:
        _LOGGER.error("No spotify oauth info: " + str(e))
        return True

    cache = hass.config.path(DEFAULT_CACHE_PATH)
    _LOGGER.info("take gate_id")
    gate_id = ais_global.get_sercure_android_id_dom()
    _LOGGER.info("gate_id: " + str(gate_id))

    j_state = json.dumps(
        {"gate_id": gate_id, "real_ip": "real_ip_place", "flow_id": "flow_id_place"}
    )
    oauth = spotipy.oauth2.SpotifyOAuth(
        spotify_client_id,
        spotify_client_secret,
        spotify_redirect_url,
        scope=spotify_scope,
        cache_path=cache,
        state=j_state,
    )

    setUrl(oauth.get_authorize_url())
    token_info = oauth.get_cached_token()
    if not token_info:
        _LOGGER.info("no spotify token in cache;")
        if AIS_SPOTIFY_TOKEN is not None:
            with open(cache, "w") as outfile:
                json.dump(AIS_SPOTIFY_TOKEN, outfile)
            token_info = oauth.get_cached_token()

    # register services
    if not token_info:
        _LOGGER.info("no spotify token exit")
        hass.async_add_job(do_the_spotify_disco(hass))
        return True

    data = hass.data[DOMAIN] = SpotifyData(hass, oauth)

    @asyncio.coroutine
    def search(call):
        _LOGGER.info("search " + str(call))
        yield from data.process_search_async(call)

    def select_search_uri(call):
        _LOGGER.info("select_search_uri")
        data.select_search_uri(call)

    def select_track_uri(call):
        _LOGGER.info("select_track_uri")
        data.select_track_uri(call)

    def change_play_queue(call):
        _LOGGER.info("change_play_queue")
        data.change_play_queue(call)

    hass.services.async_register(DOMAIN, "search", search)
    hass.services.async_register(DOMAIN, "select_search_uri", select_search_uri)
    hass.services.async_register(DOMAIN, "select_track_uri", select_track_uri)
    hass.services.async_register(DOMAIN, "change_play_queue", change_play_queue)

    return True


async def async_setup_entry(hass, config_entry):
    """Set up spotify token as config entry."""
    # setup the Spotify
    if AIS_SPOTIFY_TOKEN is None:
        await async_setup(hass, hass.config)

    return True


async def async_unload_entry(hass, config_entry):
    """Unload a config entry."""
    _LOGGER.info("Remove the Spotify token from AIS gate and cloud")
    try:
        import os

        os.remove(hass.config.path(DEFAULT_CACHE_PATH))
        _LOGGER.info("Token from cache file removed")
    except Exception as e:
        _LOGGER.error("Error removing token cache file " + str(e))
    try:
        ws_resp = aisCloud.delete_key("spotify_token")
        key = ws_resp.json()["key"]
        _LOGGER.info("Token from AIS cloud removed " + str(key))
    except Exception as e:
        _LOGGER.error("Error removing token from cloud " + str(e))
    # setup the Spotify
    await async_setup(hass, hass.config)
    return True


class SpotifyData:
    """Representation of a Spotify browser."""

    def __init__(self, hass, oauth):
        """Initialize."""
        self.hass = hass
        self._oauth = oauth
        self._spotify = None
        self._user = None
        self._token_info = self._oauth.get_cached_token()

    def refresh_spotify_instance(self):
        """Fetch a new spotify instance."""
        import spotipy

        token_refreshed = False
        need_token = self._token_info is None or self._oauth.is_token_expired(
            self._token_info
        )
        if need_token:
            new_token = self._oauth.refresh_access_token(
                self._token_info["refresh_token"]
            )
            # skip when refresh failed
            if new_token is None:
                return
            self._token_info = new_token
            token_refreshed = True
        if token_refreshed or self._spotify is None:
            self._spotify = spotipy.Spotify(auth=self._token_info.get("access_token"))
            self._user = self._spotify.me()

    def get_list_from_results(self, results, audio_type, list_info):
        items = []
        title_prefix = ""
        item_owner_id = ""
        icon = "mdi:playlist-check"
        if audio_type == "album":
            items = results["albums"]["items"]
            title_prefix = "Album: "
            icon = "mdi:album"
        elif audio_type == "artist":
            items = results["artists"]["items"]
            title_prefix = "Wykonawca: "
            icon = "mdi:artist"
        elif audio_type == "playlist":
            items = results["playlists"]["items"]
            title_prefix = "Playlista: "
            icon = "mdi:playlist-music"
        list_idx = len(list_info)
        for item in items:
            try:
                if audio_type == "playlist":
                    item_owner_id = item["owner"]["id"]
                    i_total = item["tracks"]["total"]
                elif audio_type == "artist":
                    i_total = item["popularity"]
                elif audio_type == "album":
                    i_total = item["total_tracks"]
                if i_total > 0:
                    if len(item["images"]) > 0:
                        thumbnail = item["images"][0]["url"]
                    else:
                        thumbnail = "/static/icons/favicon-100x100.png"
                    list_info[list_idx] = {
                        "uri": item["uri"],
                        "title": title_prefix + item["name"],
                        "name": title_prefix + item["name"],
                        "type": audio_type,
                        "item_owner_id": item_owner_id,
                        "thumbnail": thumbnail,
                        "icon": icon,
                        "audio_type": ais_global.G_AN_SPOTIFY_SEARCH,
                    }
                    list_idx = list_idx + 1
            except Exception as e:
                _LOGGER.error(
                    "get_list_from_results " + str(item) + " ERROR: " + str(e)
                )

        return list_info

    @asyncio.coroutine
    def get_tracks_list_async(self, item_uri, item_type, item_owner_id, item_image_url):
        items_info = {}
        idx = 0
        if item_type == "album":
            response = self._spotify.album_tracks(item_uri)
            for track in response["items"]:
                items_info[idx] = {}
                items_info[idx]["title"] = track["name"]
                items_info[idx]["name"] = track["name"]
                if item_image_url is not None:
                    items_info[idx]["thumbnail"] = item_image_url
                else:
                    items_info[idx]["thumbnail"] = "/static/icons/favicon-100x100.png"
                items_info[idx]["uri"] = track["uri"]
                items_info[idx]["audio_type"] = ais_global.G_AN_SPOTIFY
                items_info[idx]["type"] = track["type"]
                items_info[idx]["icon"] = "mdi:play"
                idx = idx + 1
        elif item_type == "artist":
            response = self._spotify.artist_top_tracks(item_uri)
            for track in response["tracks"]:
                items_info[idx] = {}
                items_info[idx]["title"] = track["name"]
                items_info[idx]["name"] = track["name"]
                if len(track["album"]["images"]) > 0:
                    items_info[idx]["thumbnail"] = track["album"]["images"][0]["url"]
                else:
                    items_info[idx]["thumbnail"] = "/static/icons/favicon-100x100.png"
                items_info[idx]["uri"] = track["uri"]
                items_info[idx]["audio_type"] = ais_global.G_AN_SPOTIFY
                items_info[idx]["type"] = track["type"]
                items_info[idx]["icon"] = "mdi:play"
                idx = idx + 1
        else:
            response = self._spotify.user_playlist(item_owner_id, item_uri)
            for items in response["tracks"]["items"]:
                items_info[idx] = {}
                items_info[idx]["title"] = items["track"]["name"]
                items_info[idx]["name"] = items["track"]["name"]
                if len(items["track"]["album"]["images"]) > 0:
                    items_info[idx]["thumbnail"] = items["track"]["album"]["images"][0][
                        "url"
                    ]
                else:
                    items_info[idx]["thumbnail"] = "/static/icons/favicon-100x100.png"
                items_info[idx]["uri"] = items["track"]["uri"]
                items_info[idx]["audio_type"] = ais_global.G_AN_SPOTIFY
                items_info[idx]["type"] = items["track"]["type"]
                items_info[idx]["icon"] = "mdi:play"
                idx = idx + 1

        # update list
        self.hass.states.async_set("sensor.spotifylist", 0, items_info)

        # from remote
        # import homeassistant.components.ais_ai_service as ais_ai
        # if ais_ai.CURR_ENTITIE == 'sensor.spotifysearchlist' and ais_ai.CURR_BUTTON_CODE == 23:
        #     ais_ai.set_curr_entity(self.hass, 'sensor.spotifylist')
        #     ais_ai.CURR_ENTITIE_ENTERED = True
        #     text = "Mamy %s utworów na liście, wybierz który mam włączyć" % (str(len(items_info)))
        #     yield from self.hass.services.async_call('ais_ai_service', 'say_it', {"text": text})
        # else:
        #     # play the first one
        #     yield from self.hass.services.async_call('ais_spotify_service', 'select_track_uri', {"id": 0})

    @asyncio.coroutine
    def process_search_async(self, call):
        """Search album on Spotify."""
        search_text = None
        if "query" in call.data:
            search_text = call.data["query"]
        if search_text is None or len(search_text.strip()) == 0:
            # get tracks from favorites
            yield from self.hass.services.async_call(
                "ais_bookmarks",
                "get_favorites",
                {"audio_source": ais_global.G_AN_SPOTIFY},
            )
            return

        self.refresh_spotify_instance()

        # Don't true search when token is expired
        if self._oauth.is_token_expired(self._token_info):
            _LOGGER.warning("Spotify failed to update, token expired.")
            return

        list_info = {}
        # artist
        results = self._spotify.search(
            q="artist:" + search_text, type="artist", limit=6
        )
        list_info = self.get_list_from_results(results, "artist", list_info)
        # album
        results = self._spotify.search(q="album:" + search_text, type="album", limit=6)
        list_info = self.get_list_from_results(results, "album", list_info)
        # playlist
        results = self._spotify.search(
            q="playlist:" + search_text, type="playlist", limit=6
        )
        list_info = self.get_list_from_results(results, "playlist", list_info)

        # update lists
        self.hass.states.async_set("sensor.spotifysearchlist", -1, list_info)
        self.hass.states.async_set("sensor.spotifylist", -1, {})

        if len(list_info) > 0:
            text = "Znaleziono: %s, włączam utwory %s" % (
                str(len(list_info)),
                list_info[0]["title"],
            )
            yield from self.hass.services.async_call(
                "ais_spotify_service", "select_search_uri", {"id": 0}
            )
        else:
            text = "Brak wyników na Spotify dla zapytania %s" % search_text
        yield from self.hass.services.async_call(
            "ais_ai_service", "say_it", {"text": text}
        )

    def select_search_uri(self, call):
        _LOGGER.info("select_search_uri")
        import json

        call_id = call.data["id"]
        state = self.hass.states.get("sensor.spotifysearchlist")
        attr = state.attributes
        track = attr.get(int(call_id))
        # update search list
        self.hass.states.async_set("sensor.spotifysearchlist", call_id, attr)

        # play the uri
        _audio_info = json.dumps(
            {
                "IMAGE_URL": track["thumbnail"],
                "NAME": track["title"],
                "MEDIA_SOURCE": ais_global.G_AN_SPOTIFY,
                "media_content_id": track["uri"],
            }
        )
        self.hass.services.call(
            "media_player",
            "play_media",
            {
                "entity_id": ais_global.G_LOCAL_EXO_PLAYER_ENTITY_ID,
                "media_content_type": "ais_content_info",
                "media_content_id": _audio_info,
            },
        )

        # get track list
        return run_coroutine_threadsafe(
            self.get_tracks_list_async(
                track["uri"], track["type"], track["item_owner_id"], track["thumbnail"]
            ),
            self.hass.loop,
        ).result()

    def select_track_uri(self, call):
        import json

        _LOGGER.info("select_track_uri")
        # """play track by id on sensor list."""
        call_id = call.data["id"]
        state = self.hass.states.get("sensor.spotifylist")
        attr = state.attributes
        track = attr.get(int(call_id))

        # update list
        self.hass.states.async_set("sensor.spotifylist", call_id, attr)
        # set stream url, image and title
        _audio_info = json.dumps(
            {
                "IMAGE_URL": track["thumbnail"],
                "NAME": track["title"],
                "MEDIA_SOURCE": ais_global.G_AN_SPOTIFY,
                "media_content_id": track["uri"],
            }
        )
        self.hass.services.call(
            "media_player",
            "play_media",
            {
                "entity_id": ais_global.G_LOCAL_EXO_PLAYER_ENTITY_ID,
                "media_content_type": "ais_content_info",
                "media_content_id": _audio_info,
            },
        )

    def change_play_queue(self, call):
        # info from android app
        _LOGGER.info("change_play_queue")



"""Constants for IPMA component."""
import logging

from homeassistant.components.weather import DOMAIN as WEATHER_DOMAIN

DOMAIN = "ipma"

HOME_LOCATION_NAME = "Home"

ENTITY_ID_SENSOR_FORMAT = WEATHER_DOMAIN + ".ipma_{}"
ENTITY_ID_SENSOR_FORMAT_HOME = ENTITY_ID_SENSOR_FORMAT.format(HOME_LOCATION_NAME)

_LOGGER = logging.getLogger(".")



"""Decorators for ZHA core registries."""
from typing import Callable, TypeVar, Union

CALLABLE_T = TypeVar("CALLABLE_T", bound=Callable)  # noqa pylint: disable=invalid-name


class DictRegistry(dict):
    """Dict Registry of items."""

    def register(
        self, name: Union[int, str], item: Union[str, CALLABLE_T] = None
    ) -> Callable[[CALLABLE_T], CALLABLE_T]:
        """Return decorator to register item with a specific name."""

        def decorator(channel: CALLABLE_T) -> CALLABLE_T:
            """Register decorated channel or item."""
            if item is None:
                self[name] = channel
            else:
                self[name] = item
            return channel

        return decorator


class SetRegistry(set):
    """Set Registry of items."""

    def register(self, name: Union[int, str]) -> Callable[[CALLABLE_T], CALLABLE_T]:
        """Return decorator to register item with a specific name."""

        def decorator(channel: CALLABLE_T) -> CALLABLE_T:
            """Register decorated channel or item."""
            self.add(name)
            return channel

        return decorator



"""
Manufacturer specific channels module for Zigbee Home Automation.

For more details about this component, please refer to the documentation at
https://home-assistant.io/components/zha/
"""
import logging

from . import AttributeListeningChannel
from .. import registries
from ..const import REPORT_CONFIG_ASAP, REPORT_CONFIG_MAX_INT, REPORT_CONFIG_MIN_INT


_LOGGER = logging.getLogger(__name__)


@registries.ZIGBEE_CHANNEL_REGISTRY.register(registries.SMARTTHINGS_HUMIDITY_CLUSTER)
class SmartThingsHumidity(AttributeListeningChannel):
    """Smart Things Humidity channel."""

    REPORT_CONFIG = [
        {
            "attr": "measured_value",
            "config": (REPORT_CONFIG_MIN_INT, REPORT_CONFIG_MAX_INT, 50),
        }
    ]


@registries.ZIGBEE_CHANNEL_REGISTRY.register(
    registries.SMARTTHINGS_ACCELERATION_CLUSTER
)
class SmartThingsAcceleration(AttributeListeningChannel):
    """Smart Things Acceleration channel."""

    REPORT_CONFIG = [
        {"attr": "acceleration", "config": REPORT_CONFIG_ASAP},
        {"attr": "x_axis", "config": REPORT_CONFIG_ASAP},
        {"attr": "y_axis", "config": REPORT_CONFIG_ASAP},
        {"attr": "z_axis", "config": REPORT_CONFIG_ASAP},
    ]



"""Support for water heater devices."""
from datetime import timedelta
import logging
import functools as ft

import voluptuous as vol

from homeassistant.helpers.temperature import display_temp as show_temp
from homeassistant.util.temperature import convert as convert_temperature
from homeassistant.helpers.entity_component import EntityComponent
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.config_validation import (  # noqa
    PLATFORM_SCHEMA,
    PLATFORM_SCHEMA_BASE,
)
import homeassistant.helpers.config_validation as cv
from homeassistant.const import (
    ATTR_ENTITY_ID,
    ATTR_TEMPERATURE,
    SERVICE_TURN_ON,
    SERVICE_TURN_OFF,
    STATE_ON,
    STATE_OFF,
    TEMP_CELSIUS,
    PRECISION_WHOLE,
    PRECISION_TENTHS,
    TEMP_FAHRENHEIT,
)


# mypy: allow-untyped-defs, no-check-untyped-defs

DEFAULT_MIN_TEMP = 110
DEFAULT_MAX_TEMP = 140

DOMAIN = "water_heater"

ENTITY_ID_FORMAT = DOMAIN + ".{}"
SCAN_INTERVAL = timedelta(seconds=60)

SERVICE_SET_AWAY_MODE = "set_away_mode"
SERVICE_SET_TEMPERATURE = "set_temperature"
SERVICE_SET_OPERATION_MODE = "set_operation_mode"

STATE_ECO = "eco"
STATE_ELECTRIC = "electric"
STATE_PERFORMANCE = "performance"
STATE_HIGH_DEMAND = "high_demand"
STATE_HEAT_PUMP = "heat_pump"
STATE_GAS = "gas"

SUPPORT_TARGET_TEMPERATURE = 1
SUPPORT_OPERATION_MODE = 2
SUPPORT_AWAY_MODE = 4

ATTR_MAX_TEMP = "max_temp"
ATTR_MIN_TEMP = "min_temp"
ATTR_AWAY_MODE = "away_mode"
ATTR_OPERATION_MODE = "operation_mode"
ATTR_OPERATION_LIST = "operation_list"
ATTR_TARGET_TEMP_HIGH = "target_temp_high"
ATTR_TARGET_TEMP_LOW = "target_temp_low"
ATTR_CURRENT_TEMPERATURE = "current_temperature"

CONVERTIBLE_ATTRIBUTE = [ATTR_TEMPERATURE]

_LOGGER = logging.getLogger(__name__)

ON_OFF_SERVICE_SCHEMA = vol.Schema({vol.Optional(ATTR_ENTITY_ID): cv.comp_entity_ids})

SET_AWAY_MODE_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_ENTITY_ID): cv.comp_entity_ids,
        vol.Required(ATTR_AWAY_MODE): cv.boolean,
    }
)
SET_TEMPERATURE_SCHEMA = vol.Schema(
    vol.All(
        {
            vol.Required(ATTR_TEMPERATURE, "temperature"): vol.Coerce(float),
            vol.Optional(ATTR_ENTITY_ID): cv.comp_entity_ids,
            vol.Optional(ATTR_OPERATION_MODE): cv.string,
        }
    )
)
SET_OPERATION_MODE_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_ENTITY_ID): cv.comp_entity_ids,
        vol.Required(ATTR_OPERATION_MODE): cv.string,
    }
)


async def async_setup(hass, config):
    """Set up water_heater devices."""
    component = hass.data[DOMAIN] = EntityComponent(
        _LOGGER, DOMAIN, hass, SCAN_INTERVAL
    )
    await component.async_setup(config)

    component.async_register_entity_service(
        SERVICE_SET_AWAY_MODE, SET_AWAY_MODE_SCHEMA, async_service_away_mode
    )
    component.async_register_entity_service(
        SERVICE_SET_TEMPERATURE, SET_TEMPERATURE_SCHEMA, async_service_temperature_set
    )
    component.async_register_entity_service(
        SERVICE_SET_OPERATION_MODE,
        SET_OPERATION_MODE_SCHEMA,
        "async_set_operation_mode",
    )
    component.async_register_entity_service(
        SERVICE_TURN_OFF, ON_OFF_SERVICE_SCHEMA, "async_turn_off"
    )
    component.async_register_entity_service(
        SERVICE_TURN_ON, ON_OFF_SERVICE_SCHEMA, "async_turn_on"
    )

    return True


async def async_setup_entry(hass, entry):
    """Set up a config entry."""
    return await hass.data[DOMAIN].async_setup_entry(entry)


async def async_unload_entry(hass, entry):
    """Unload a config entry."""
    return await hass.data[DOMAIN].async_unload_entry(entry)


class WaterHeaterDevice(Entity):
    """Representation of a water_heater device."""

    @property
    def state(self):
        """Return the current state."""
        return self.current_operation

    @property
    def precision(self):
        """Return the precision of the system."""
        if self.hass.config.units.temperature_unit == TEMP_CELSIUS:
            return PRECISION_TENTHS
        return PRECISION_WHOLE

    @property
    def state_attributes(self):
        """Return the optional state attributes."""
        data = {
            ATTR_CURRENT_TEMPERATURE: show_temp(
                self.hass,
                self.current_temperature,
                self.temperature_unit,
                self.precision,
            ),
            ATTR_MIN_TEMP: show_temp(
                self.hass, self.min_temp, self.temperature_unit, self.precision
            ),
            ATTR_MAX_TEMP: show_temp(
                self.hass, self.max_temp, self.temperature_unit, self.precision
            ),
            ATTR_TEMPERATURE: show_temp(
                self.hass,
                self.target_temperature,
                self.temperature_unit,
                self.precision,
            ),
            ATTR_TARGET_TEMP_HIGH: show_temp(
                self.hass,
                self.target_temperature_high,
                self.temperature_unit,
                self.precision,
            ),
            ATTR_TARGET_TEMP_LOW: show_temp(
                self.hass,
                self.target_temperature_low,
                self.temperature_unit,
                self.precision,
            ),
        }

        supported_features = self.supported_features

        if supported_features & SUPPORT_OPERATION_MODE:
            data[ATTR_OPERATION_MODE] = self.current_operation
            if self.operation_list:
                data[ATTR_OPERATION_LIST] = self.operation_list

        if supported_features & SUPPORT_AWAY_MODE:
            is_away = self.is_away_mode_on
            data[ATTR_AWAY_MODE] = STATE_ON if is_away else STATE_OFF

        return data

    @property
    def temperature_unit(self):
        """Return the unit of measurement used by the platform."""
        raise NotImplementedError

    @property
    def current_operation(self):
        """Return current operation ie. eco, electric, performance, ..."""
        return None

    @property
    def operation_list(self):
        """Return the list of available operation modes."""
        return None

    @property
    def current_temperature(self):
        """Return the current temperature."""
        return None

    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        return None

    @property
    def target_temperature_high(self):
        """Return the highbound target temperature we try to reach."""
        return None

    @property
    def target_temperature_low(self):
        """Return the lowbound target temperature we try to reach."""
        return None

    @property
    def is_away_mode_on(self):
        """Return true if away mode is on."""
        return None

    def set_temperature(self, **kwargs):
        """Set new target temperature."""
        raise NotImplementedError()

    async def async_set_temperature(self, **kwargs):
        """Set new target temperature."""
        await self.hass.async_add_executor_job(
            ft.partial(self.set_temperature, **kwargs)
        )

    def set_operation_mode(self, operation_mode):
        """Set new target operation mode."""
        raise NotImplementedError()

    async def async_set_operation_mode(self, operation_mode):
        """Set new target operation mode."""
        await self.hass.async_add_executor_job(self.set_operation_mode, operation_mode)

    def turn_away_mode_on(self):
        """Turn away mode on."""
        raise NotImplementedError()

    async def async_turn_away_mode_on(self):
        """Turn away mode on."""
        await self.hass.async_add_executor_job(self.turn_away_mode_on)

    def turn_away_mode_off(self):
        """Turn away mode off."""
        raise NotImplementedError()

    async def async_turn_away_mode_off(self):
        """Turn away mode off."""
        await self.hass.async_add_executor_job(self.turn_away_mode_off)

    @property
    def supported_features(self):
        """Return the list of supported features."""
        raise NotImplementedError()

    @property
    def min_temp(self):
        """Return the minimum temperature."""
        return convert_temperature(
            DEFAULT_MIN_TEMP, TEMP_FAHRENHEIT, self.temperature_unit
        )

    @property
    def max_temp(self):
        """Return the maximum temperature."""
        return convert_temperature(
            DEFAULT_MAX_TEMP, TEMP_FAHRENHEIT, self.temperature_unit
        )


async def async_service_away_mode(entity, service):
    """Handle away mode service."""
    if service.data[ATTR_AWAY_MODE]:
        await entity.async_turn_away_mode_on()
    else:
        await entity.async_turn_away_mode_off()


async def async_service_temperature_set(entity, service):
    """Handle set temperature service."""
    hass = entity.hass
    kwargs = {}

    for value, temp in service.data.items():
        if value in CONVERTIBLE_ATTRIBUTE:
            kwargs[value] = convert_temperature(
                temp, hass.config.units.temperature_unit, entity.temperature_unit
            )
        else:
            kwargs[value] = temp

    await entity.async_set_temperature(**kwargs)



"""Provide animated GIF loops of BOM radar imagery."""
import voluptuous as vol

from homeassistant.components.camera import PLATFORM_SCHEMA, Camera
from homeassistant.const import CONF_ID, CONF_NAME
from homeassistant.helpers import config_validation as cv

CONF_DELTA = "delta"
CONF_FRAMES = "frames"
CONF_LOCATION = "location"
CONF_OUTFILE = "filename"

LOCATIONS = [
    "Adelaide",
    "Albany",
    "AliceSprings",
    "Bairnsdale",
    "Bowen",
    "Brisbane",
    "Broome",
    "Cairns",
    "Canberra",
    "Carnarvon",
    "Ceduna",
    "Dampier",
    "Darwin",
    "Emerald",
    "Esperance",
    "Geraldton",
    "Giles",
    "Gladstone",
    "Gove",
    "Grafton",
    "Gympie",
    "HallsCreek",
    "Hobart",
    "Kalgoorlie",
    "Katherine",
    "Learmonth",
    "Longreach",
    "Mackay",
    "Marburg",
    "Melbourne",
    "Mildura",
    "Moree",
    "MorningtonIs",
    "MountIsa",
    "MtGambier",
    "Namoi",
    "Newcastle",
    "Newdegate",
    "NorfolkIs",
    "NWTasmania",
    "Perth",
    "PortHedland",
    "SellicksHill",
    "SouthDoodlakine",
    "Sydney",
    "Townsville",
    "WaggaWagga",
    "Warrego",
    "Warruwi",
    "Watheroo",
    "Weipa",
    "WillisIs",
    "Wollongong",
    "Woomera",
    "Wyndham",
    "Yarrawonga",
]


def _validate_schema(config):
    if config.get(CONF_LOCATION) is None:
        if not all(config.get(x) for x in (CONF_ID, CONF_DELTA, CONF_FRAMES)):
            raise vol.Invalid(
                "Specify '{}', '{}' and '{}' when '{}' is unspecified".format(
                    CONF_ID, CONF_DELTA, CONF_FRAMES, CONF_LOCATION
                )
            )
    return config


LOCATIONS_MSG = "Set '{}' to one of: {}".format(
    CONF_LOCATION, ", ".join(sorted(LOCATIONS))
)
XOR_MSG = "Specify exactly one of '{}' or '{}'".format(CONF_ID, CONF_LOCATION)

PLATFORM_SCHEMA = vol.All(
    PLATFORM_SCHEMA.extend(
        {
            vol.Exclusive(CONF_ID, "xor", msg=XOR_MSG): cv.string,
            vol.Exclusive(CONF_LOCATION, "xor", msg=XOR_MSG): vol.In(
                LOCATIONS, msg=LOCATIONS_MSG
            ),
            vol.Optional(CONF_DELTA): cv.positive_int,
            vol.Optional(CONF_FRAMES): cv.positive_int,
            vol.Optional(CONF_NAME): cv.string,
            vol.Optional(CONF_OUTFILE): cv.string,
        }
    ),
    _validate_schema,
)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up BOM radar-loop camera component."""
    location = config.get(CONF_LOCATION) or "ID {}".format(config.get(CONF_ID))
    name = config.get(CONF_NAME) or "BOM Radar Loop - {}".format(location)
    args = [
        config.get(x)
        for x in (CONF_LOCATION, CONF_ID, CONF_DELTA, CONF_FRAMES, CONF_OUTFILE)
    ]
    add_entities([BOMRadarCam(name, *args)])


class BOMRadarCam(Camera):
    """A camera component producing animated BOM radar-imagery GIFs."""

    def __init__(self, name, location, radar_id, delta, frames, outfile):
        """Initialize the component."""
        from bomradarloop import BOMRadarLoop

        super().__init__()
        self._name = name
        self._cam = BOMRadarLoop(location, radar_id, delta, frames, outfile)

    def camera_image(self):
        """Return the current BOM radar-loop image."""
        return self._cam.current

    @property
    def name(self):
        """Return the component name."""
        return self._name



"""Support the ElkM1 Gold and ElkM1 EZ8 alarm/integration panels."""
import logging
import re

import voluptuous as vol
from homeassistant.const import (
    CONF_EXCLUDE,
    CONF_HOST,
    CONF_INCLUDE,
    CONF_PASSWORD,
    CONF_TEMPERATURE_UNIT,
    CONF_USERNAME,
)
from homeassistant.core import HomeAssistant, callback  # noqa
from homeassistant.helpers import config_validation as cv
from homeassistant.helpers import discovery
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.typing import ConfigType  # noqa

DOMAIN = "elkm1"

CONF_AREA = "area"
CONF_COUNTER = "counter"
CONF_ENABLED = "enabled"
CONF_KEYPAD = "keypad"
CONF_OUTPUT = "output"
CONF_PLC = "plc"
CONF_SETTING = "setting"
CONF_TASK = "task"
CONF_THERMOSTAT = "thermostat"
CONF_ZONE = "zone"
CONF_PREFIX = "prefix"

_LOGGER = logging.getLogger(__name__)

SUPPORTED_DOMAINS = [
    "alarm_control_panel",
    "climate",
    "light",
    "scene",
    "sensor",
    "switch",
]

SPEAK_SERVICE_SCHEMA = vol.Schema(
    {
        vol.Required("number"): vol.All(vol.Coerce(int), vol.Range(min=0, max=999)),
        vol.Optional("prefix", default=""): cv.string,
    }
)


def _host_validator(config):
    """Validate that a host is properly configured."""
    if config[CONF_HOST].startswith("elks://"):
        if CONF_USERNAME not in config or CONF_PASSWORD not in config:
            raise vol.Invalid("Specify username and password for elks://")
    elif not config[CONF_HOST].startswith("elk://") and not config[
        CONF_HOST
    ].startswith("serial://"):
        raise vol.Invalid("Invalid host URL")
    return config


def _elk_range_validator(rng):
    def _housecode_to_int(val):
        match = re.search(r"^([a-p])(0[1-9]|1[0-6]|[1-9])$", val.lower())
        if match:
            return (ord(match.group(1)) - ord("a")) * 16 + int(match.group(2))
        raise vol.Invalid("Invalid range")

    def _elk_value(val):
        return int(val) if val.isdigit() else _housecode_to_int(val)

    vals = [s.strip() for s in str(rng).split("-")]
    start = _elk_value(vals[0])
    end = start if len(vals) == 1 else _elk_value(vals[1])
    return (start, end)


def _has_all_unique_prefixes(value):
    """Validate that each m1 configured has a unique prefix.

    Uniqueness is determined case-independently.
    """
    prefixes = [device[CONF_PREFIX] for device in value]
    schema = vol.Schema(vol.Unique())
    schema(prefixes)
    return value


DEVICE_SCHEMA_SUBDOMAIN = vol.Schema(
    {
        vol.Optional(CONF_ENABLED, default=True): cv.boolean,
        vol.Optional(CONF_INCLUDE, default=[]): [_elk_range_validator],
        vol.Optional(CONF_EXCLUDE, default=[]): [_elk_range_validator],
    }
)

DEVICE_SCHEMA = vol.Schema(
    {
        vol.Required(CONF_HOST): cv.string,
        vol.Optional(CONF_PREFIX, default=""): vol.All(cv.string, vol.Lower),
        vol.Optional(CONF_USERNAME, default=""): cv.string,
        vol.Optional(CONF_PASSWORD, default=""): cv.string,
        vol.Optional(CONF_TEMPERATURE_UNIT, default="F"): cv.temperature_unit,
        vol.Optional(CONF_AREA, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_COUNTER, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_KEYPAD, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_OUTPUT, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_PLC, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_SETTING, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_TASK, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_THERMOSTAT, default={}): DEVICE_SCHEMA_SUBDOMAIN,
        vol.Optional(CONF_ZONE, default={}): DEVICE_SCHEMA_SUBDOMAIN,
    },
    _host_validator,
)

CONFIG_SCHEMA = vol.Schema(
    {DOMAIN: vol.All(cv.ensure_list, [DEVICE_SCHEMA], _has_all_unique_prefixes)},
    extra=vol.ALLOW_EXTRA,
)


async def async_setup(hass: HomeAssistant, hass_config: ConfigType) -> bool:
    """Set up the Elk M1 platform."""
    from elkm1_lib.const import Max
    import elkm1_lib as elkm1

    devices = {}
    elk_datas = {}

    configs = {
        CONF_AREA: Max.AREAS.value,
        CONF_COUNTER: Max.COUNTERS.value,
        CONF_KEYPAD: Max.KEYPADS.value,
        CONF_OUTPUT: Max.OUTPUTS.value,
        CONF_PLC: Max.LIGHTS.value,
        CONF_SETTING: Max.SETTINGS.value,
        CONF_TASK: Max.TASKS.value,
        CONF_THERMOSTAT: Max.THERMOSTATS.value,
        CONF_ZONE: Max.ZONES.value,
    }

    def _included(ranges, set_to, values):
        for rng in ranges:
            if not rng[0] <= rng[1] <= len(values):
                raise vol.Invalid("Invalid range {}".format(rng))
            values[rng[0] - 1 : rng[1]] = [set_to] * (rng[1] - rng[0] + 1)

    for index, conf in enumerate(hass_config[DOMAIN]):
        _LOGGER.debug("Setting up elkm1 #%d - %s", index, conf["host"])

        config = {"temperature_unit": conf[CONF_TEMPERATURE_UNIT]}
        config["panel"] = {"enabled": True, "included": [True]}

        for item, max_ in configs.items():
            config[item] = {
                "enabled": conf[item][CONF_ENABLED],
                "included": [not conf[item]["include"]] * max_,
            }
            try:
                _included(conf[item]["include"], True, config[item]["included"])
                _included(conf[item]["exclude"], False, config[item]["included"])
            except (ValueError, vol.Invalid) as err:
                _LOGGER.error("Config item: %s; %s", item, err)
                return False

        prefix = conf[CONF_PREFIX]
        elk = elkm1.Elk(
            {
                "url": conf[CONF_HOST],
                "userid": conf[CONF_USERNAME],
                "password": conf[CONF_PASSWORD],
            }
        )
        elk.connect()

        devices[prefix] = elk
        elk_datas[prefix] = {
            "elk": elk,
            "prefix": prefix,
            "config": config,
            "keypads": {},
        }

    _create_elk_services(hass, devices)

    hass.data[DOMAIN] = elk_datas
    for component in SUPPORTED_DOMAINS:
        hass.async_create_task(
            discovery.async_load_platform(hass, component, DOMAIN, {}, hass_config)
        )

    return True


def _create_elk_services(hass, elks):
    def _speak_word_service(service):
        prefix = service.data["prefix"]
        elk = elks.get(prefix)
        if elk is None:
            _LOGGER.error("No elk m1 with prefix for speak_word: '%s'", prefix)
            return
        elk.panel.speak_word(service.data["number"])

    def _speak_phrase_service(service):
        prefix = service.data["prefix"]
        elk = elks.get(prefix)
        if elk is None:
            _LOGGER.error("No elk m1 with prefix for speak_phrase: '%s'", prefix)
            return
        elk.panel.speak_phrase(service.data["number"])

    hass.services.async_register(
        DOMAIN, "speak_word", _speak_word_service, SPEAK_SERVICE_SCHEMA
    )
    hass.services.async_register(
        DOMAIN, "speak_phrase", _speak_phrase_service, SPEAK_SERVICE_SCHEMA
    )


def create_elk_entities(elk_data, elk_elements, element_type, class_, entities):
    """Create the ElkM1 devices of a particular class."""
    if elk_data["config"][element_type]["enabled"]:
        elk = elk_data["elk"]
        _LOGGER.debug("Creating elk entities for %s", elk)
        for element in elk_elements:
            if elk_data["config"][element_type]["included"][element.index]:
                entities.append(class_(element, elk, elk_data))
    return entities


class ElkEntity(Entity):
    """Base class for all Elk entities."""

    def __init__(self, element, elk, elk_data):
        """Initialize the base of all Elk devices."""
        self._elk = elk
        self._element = element
        self._prefix = elk_data["prefix"]
        self._temperature_unit = elk_data["config"]["temperature_unit"]
        # unique_id starts with elkm1_ iff there is no prefix
        # it starts with elkm1m_{prefix} iff there is a prefix
        # this is to avoid a conflict between
        # prefix=foo, name=bar  (which would be elkm1_foo_bar)
        #   - and -
        # prefix="", name="foo bar" (which would be elkm1_foo_bar also)
        # we could have used elkm1__foo_bar for the latter, but that
        # would have been a breaking change
        if self._prefix != "":
            uid_start = "elkm1m_{prefix}".format(prefix=self._prefix)
        else:
            uid_start = "elkm1"
        self._unique_id = "{uid_start}_{name}".format(
            uid_start=uid_start, name=self._element.default_name("_")
        ).lower()

    @property
    def name(self):
        """Name of the element."""
        return "{p}{n}".format(p=self._prefix, n=self._element.name)

    @property
    def unique_id(self):
        """Return unique id of the element."""
        return self._unique_id

    @property
    def should_poll(self) -> bool:
        """Don't poll this device."""
        return False

    @property
    def device_state_attributes(self):
        """Return the default attributes of the element."""
        return {**self._element.as_dict(), **self.initial_attrs()}

    @property
    def available(self):
        """Is the entity available to be updated."""
        return self._elk.is_connected()

    def initial_attrs(self):
        """Return the underlying element's attributes as a dict."""
        attrs = {}
        attrs["index"] = self._element.index + 1
        return attrs

    def _element_changed(self, element, changeset):
        pass

    @callback
    def _element_callback(self, element, changeset):
        """Handle callback from an Elk element that has changed."""
        self._element_changed(element, changeset)
        self.async_schedule_update_ha_state(True)

    async def async_added_to_hass(self):
        """Register callback for ElkM1 changes and update entity state."""
        self._element.add_callback(self._element_callback)
        self._element_callback(self._element, {})



"""Config flow to configure deCONZ component."""
import asyncio
from copy import copy

import async_timeout
import voluptuous as vol

from pydeconz.errors import ResponseError, RequestError
from pydeconz.utils import async_discovery, async_get_api_key, async_get_bridgeid

from homeassistant import config_entries
from homeassistant.const import CONF_API_KEY, CONF_HOST, CONF_PORT
from homeassistant.core import callback
from homeassistant.helpers import aiohttp_client

from .const import (
    CONF_ALLOW_CLIP_SENSOR,
    CONF_ALLOW_DECONZ_GROUPS,
    CONF_BRIDGEID,
    DEFAULT_PORT,
    DOMAIN,
)

DECONZ_MANUFACTURERURL = "http://www.dresden-elektronik.de"
CONF_SERIAL = "serial"
ATTR_UUID = "udn"


@callback
def configured_gateways(hass):
    """Return a set of all configured gateways."""
    return {
        entry.data[CONF_BRIDGEID]: entry
        for entry in hass.config_entries.async_entries(DOMAIN)
    }


@callback
def get_master_gateway(hass):
    """Return the gateway which is marked as master."""
    for gateway in hass.data[DOMAIN].values():
        if gateway.master:
            return gateway


@config_entries.HANDLERS.register(DOMAIN)
class DeconzFlowHandler(config_entries.ConfigFlow):
    """Handle a deCONZ config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_PUSH

    _hassio_discovery = None

    @staticmethod
    @callback
    def async_get_options_flow(config_entry):
        """Get the options flow for this handler."""
        return DeconzOptionsFlowHandler(config_entry)

    def __init__(self):
        """Initialize the deCONZ config flow."""
        self.bridges = []
        self.deconz_config = {}

    async def async_step_init(self, user_input=None):
        """Needed in order to not require re-translation of strings."""
        return await self.async_step_user(user_input)

    async def async_step_user(self, user_input=None):
        """Handle a deCONZ config flow start.

        If only one bridge is found go to link step.
        If more than one bridge is found let user choose bridge to link.
        If no bridge is found allow user to manually input configuration.
        """
        if user_input is not None:
            for bridge in self.bridges:
                if bridge[CONF_HOST] == user_input[CONF_HOST]:
                    self.deconz_config = bridge
                    return await self.async_step_link()

            self.deconz_config = user_input
            return await self.async_step_link()

        session = aiohttp_client.async_get_clientsession(self.hass)

        try:
            with async_timeout.timeout(10):
                self.bridges = await async_discovery(session)

        except asyncio.TimeoutError:
            self.bridges = []

        if len(self.bridges) == 1:
            self.deconz_config = self.bridges[0]
            return await self.async_step_link()

        if len(self.bridges) > 1:
            hosts = []

            for bridge in self.bridges:
                hosts.append(bridge[CONF_HOST])

            return self.async_show_form(
                step_id="init",
                data_schema=vol.Schema({vol.Required(CONF_HOST): vol.In(hosts)}),
            )

        return self.async_show_form(
            step_id="init",
            data_schema=vol.Schema(
                {
                    vol.Required(CONF_HOST): str,
                    vol.Required(CONF_PORT, default=DEFAULT_PORT): int,
                }
            ),
        )

    async def async_step_link(self, user_input=None):
        """Attempt to link with the deCONZ bridge."""
        errors = {}

        if user_input is not None:
            session = aiohttp_client.async_get_clientsession(self.hass)

            try:
                with async_timeout.timeout(10):
                    api_key = await async_get_api_key(session, **self.deconz_config)

            except (ResponseError, RequestError, asyncio.TimeoutError):
                errors["base"] = "no_key"

            else:
                self.deconz_config[CONF_API_KEY] = api_key
                return await self._create_entry()

        return self.async_show_form(step_id="link", errors=errors)

    async def _create_entry(self):
        """Create entry for gateway."""
        if CONF_BRIDGEID not in self.deconz_config:
            session = aiohttp_client.async_get_clientsession(self.hass)

            try:
                with async_timeout.timeout(10):
                    self.deconz_config[CONF_BRIDGEID] = await async_get_bridgeid(
                        session, **self.deconz_config
                    )

            except asyncio.TimeoutError:
                return self.async_abort(reason="no_bridges")

        return self.async_create_entry(
            title="deCONZ-" + self.deconz_config[CONF_BRIDGEID], data=self.deconz_config
        )

    async def _update_entry(self, entry, host):
        """Update existing entry."""
        if entry.data[CONF_HOST] == host:
            return self.async_abort(reason="already_configured")

        entry.data[CONF_HOST] = host
        self.hass.config_entries.async_update_entry(entry)
        return self.async_abort(reason="updated_instance")

    async def async_step_ssdp(self, discovery_info):
        """Handle a discovered deCONZ bridge."""
        from homeassistant.components.ssdp import ATTR_MANUFACTURERURL, ATTR_SERIAL

        if discovery_info[ATTR_MANUFACTURERURL] != DECONZ_MANUFACTURERURL:
            return self.async_abort(reason="not_deconz_bridge")

        uuid = discovery_info[ATTR_UUID].replace("uuid:", "")
        gateways = {
            gateway.api.config.uuid: gateway
            for gateway in self.hass.data.get(DOMAIN, {}).values()
        }

        if uuid in gateways:
            entry = gateways[uuid].config_entry
            return await self._update_entry(entry, discovery_info[CONF_HOST])

        bridgeid = discovery_info[ATTR_SERIAL]
        if any(
            bridgeid == flow["context"][CONF_BRIDGEID]
            for flow in self._async_in_progress()
        ):
            return self.async_abort(reason="already_in_progress")

        # pylint: disable=unsupported-assignment-operation
        self.context[CONF_BRIDGEID] = bridgeid

        deconz_config = {
            CONF_HOST: discovery_info[CONF_HOST],
            CONF_PORT: discovery_info[CONF_PORT],
        }

        return await self.async_step_import(deconz_config)

    async def async_step_import(self, import_config):
        """Import a deCONZ bridge as a config entry.

        This flow is triggered by `async_setup` for configured bridges.
        This flow is also triggered by `async_step_discovery`.

        This will execute for any bridge that does not have a
        config entry yet (based on host).

        If an API key is provided, we will create an entry.
        Otherwise we will delegate to `link` step which
        will ask user to link the bridge.
        """
        self.deconz_config = import_config
        if CONF_API_KEY not in import_config:
            return await self.async_step_link()

        return await self._create_entry()

    async def async_step_hassio(self, user_input=None):
        """Prepare configuration for a Hass.io deCONZ bridge.

        This flow is triggered by the discovery component.
        """
        bridgeid = user_input[CONF_SERIAL]
        gateway_entries = configured_gateways(self.hass)

        if bridgeid in gateway_entries:
            entry = gateway_entries[bridgeid]
            return await self._update_entry(entry, user_input[CONF_HOST])

        self._hassio_discovery = user_input

        return await self.async_step_hassio_confirm()

    async def async_step_hassio_confirm(self, user_input=None):
        """Confirm a Hass.io discovery."""
        if user_input is not None:
            self.deconz_config = {
                CONF_HOST: self._hassio_discovery[CONF_HOST],
                CONF_PORT: self._hassio_discovery[CONF_PORT],
                CONF_BRIDGEID: self._hassio_discovery[CONF_SERIAL],
                CONF_API_KEY: self._hassio_discovery[CONF_API_KEY],
            }

            return await self._create_entry()

        return self.async_show_form(
            step_id="hassio_confirm",
            description_placeholders={"addon": self._hassio_discovery["addon"]},
        )


class DeconzOptionsFlowHandler(config_entries.OptionsFlow):
    """Handle deCONZ options."""

    def __init__(self, config_entry):
        """Initialize deCONZ options flow."""
        self.config_entry = config_entry
        self.options = copy(config_entry.options)

    async def async_step_init(self, user_input=None):
        """Manage the deCONZ options."""
        return await self.async_step_deconz_devices()

    async def async_step_deconz_devices(self, user_input=None):
        """Manage the deconz devices options."""
        if user_input is not None:
            self.options[CONF_ALLOW_CLIP_SENSOR] = user_input[CONF_ALLOW_CLIP_SENSOR]
            self.options[CONF_ALLOW_DECONZ_GROUPS] = user_input[
                CONF_ALLOW_DECONZ_GROUPS
            ]
            return self.async_create_entry(title="", data=self.options)

        return self.async_show_form(
            step_id="deconz_devices",
            data_schema=vol.Schema(
                {
                    vol.Optional(
                        CONF_ALLOW_CLIP_SENSOR,
                        default=self.config_entry.options[CONF_ALLOW_CLIP_SENSOR],
                    ): bool,
                    vol.Optional(
                        CONF_ALLOW_DECONZ_GROUPS,
                        default=self.config_entry.options[CONF_ALLOW_DECONZ_GROUPS],
                    ): bool,
                }
            ),
        )



"""Support for BT Smart Hub (Sometimes referred to as BT Home Hub 6)."""
import logging

import voluptuous as vol

import homeassistant.helpers.config_validation as cv
from homeassistant.components.device_tracker import (
    DOMAIN,
    PLATFORM_SCHEMA,
    DeviceScanner,
)
from homeassistant.const import CONF_HOST

_LOGGER = logging.getLogger(__name__)

CONF_DEFAULT_IP = "192.168.1.254"

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {vol.Optional(CONF_HOST, default=CONF_DEFAULT_IP): cv.string}
)


def get_scanner(hass, config):
    """Return a BT Smart Hub scanner if successful."""
    scanner = BTSmartHubScanner(config[DOMAIN])

    return scanner if scanner.success_init else None


class BTSmartHubScanner(DeviceScanner):
    """This class queries a BT Smart Hub."""

    def __init__(self, config):
        """Initialise the scanner."""
        _LOGGER.debug("Initialising BT Smart Hub")
        self.host = config[CONF_HOST]
        self.last_results = {}
        self.success_init = False

        # Test the router is accessible
        data = self.get_bt_smarthub_data()
        if data:
            self.success_init = True
        else:
            _LOGGER.info("Failed to connect to %s", self.host)

    def scan_devices(self):
        """Scan for new devices and return a list with found device IDs."""
        self._update_info()
        return [client["mac"] for client in self.last_results]

    def get_device_name(self, device):
        """Return the name of the given device or None if we don't know."""
        if not self.last_results:
            return None
        for client in self.last_results:
            if client["mac"] == device:
                return client["host"]
        return None

    def _update_info(self):
        """Ensure the information from the BT Smart Hub is up to date."""
        if not self.success_init:
            return

        _LOGGER.info("Scanning")
        data = self.get_bt_smarthub_data()
        if not data:
            _LOGGER.warning("Error scanning devices")
            return

        clients = [client for client in data.values()]
        self.last_results = clients

    def get_bt_smarthub_data(self):
        """Retrieve data from BT Smart Hub and return parsed result."""
        import btsmarthub_devicelist

        # Request data from bt smarthub into a list of dicts.
        data = btsmarthub_devicelist.get_devicelist(
            router_ip=self.host, only_active_devices=True
        )
        # Renaming keys from parsed result.
        devices = {}
        for device in data:
            try:
                devices[device["UserHostName"]] = {
                    "ip": device["IPAddress"],
                    "mac": device["PhysAddress"],
                    "host": device["UserHostName"],
                    "status": device["Active"],
                }
            except KeyError:
                pass
        return devices



"""Support for a Genius Hub system."""
from datetime import timedelta
import logging
from typing import Awaitable

import aiohttp
import voluptuous as vol

from geniushubclient import GeniusHub

from homeassistant.const import CONF_HOST, CONF_PASSWORD, CONF_TOKEN, CONF_USERNAME
from homeassistant.core import callback
from homeassistant.helpers import config_validation as cv
from homeassistant.helpers.aiohttp_client import async_get_clientsession
from homeassistant.helpers.discovery import async_load_platform
from homeassistant.helpers.dispatcher import (
    async_dispatcher_send,
    async_dispatcher_connect,
)
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.event import async_track_time_interval

_LOGGER = logging.getLogger(__name__)

DOMAIN = "geniushub"

SCAN_INTERVAL = timedelta(seconds=60)

_V1_API_SCHEMA = vol.Schema({vol.Required(CONF_TOKEN): cv.string})
_V3_API_SCHEMA = vol.Schema(
    {
        vol.Required(CONF_HOST): cv.string,
        vol.Required(CONF_USERNAME): cv.string,
        vol.Required(CONF_PASSWORD): cv.string,
    }
)
CONFIG_SCHEMA = vol.Schema(
    {DOMAIN: vol.Any(_V3_API_SCHEMA, _V1_API_SCHEMA)}, extra=vol.ALLOW_EXTRA
)


async def async_setup(hass, hass_config):
    """Create a Genius Hub system."""
    kwargs = dict(hass_config[DOMAIN])
    if CONF_HOST in kwargs:
        args = (kwargs.pop(CONF_HOST),)
    else:
        args = (kwargs.pop(CONF_TOKEN),)

    hass.data[DOMAIN] = {}
    broker = GeniusBroker(hass, args, kwargs)

    try:
        await broker.client.update()
    except aiohttp.ClientResponseError as err:
        _LOGGER.error("Setup failed, check your configuration, %s", err)
        return False
    broker.make_debug_log_entries()

    async_track_time_interval(hass, broker.async_update, SCAN_INTERVAL)

    for platform in ["climate", "water_heater"]:
        hass.async_create_task(
            async_load_platform(hass, platform, DOMAIN, {}, hass_config)
        )

    if broker.client.api_version == 3:  # pylint: disable=no-member
        for platform in ["sensor", "binary_sensor"]:
            hass.async_create_task(
                async_load_platform(hass, platform, DOMAIN, {}, hass_config)
            )

    return True


class GeniusBroker:
    """Container for geniushub client and data."""

    def __init__(self, hass, args, kwargs):
        """Initialize the geniushub client."""
        self.hass = hass
        self.client = hass.data[DOMAIN]["client"] = GeniusHub(
            *args, **kwargs, session=async_get_clientsession(hass)
        )

    async def async_update(self, now, **kwargs):
        """Update the geniushub client's data."""
        try:
            await self.client.update()
        except aiohttp.ClientResponseError as err:
            _LOGGER.warning("Update failed, %s", err)
            return
        self.make_debug_log_entries()

        async_dispatcher_send(self.hass, DOMAIN)

    def make_debug_log_entries(self):
        """Make any useful debug log entries."""
        # pylint: disable=protected-access
        _LOGGER.debug(
            "Raw JSON: \n\nclient._zones = %s \n\nclient._devices = %s",
            self.client._zones,
            self.client._devices,
        )


class GeniusEntity(Entity):
    """Base for all Genius Hub endtities."""

    def __init__(self):
        """Initialize the entity."""
        self._name = None

    async def async_added_to_hass(self) -> Awaitable[None]:
        """Set up a listener when this entity is added to HA."""
        async_dispatcher_connect(self.hass, DOMAIN, self._refresh)

    @callback
    def _refresh(self) -> None:
        self.async_schedule_update_ha_state(force_refresh=True)

    @property
    def name(self) -> str:
        """Return the name of the geniushub entity."""
        return self._name

    @property
    def should_poll(self) -> bool:
        """Return False as geniushub entities should not be polled."""
        return False



"""Support for the GPSLogger device tracking."""
import logging

from homeassistant.core import callback
from homeassistant.const import (
    ATTR_BATTERY_LEVEL,
    ATTR_GPS_ACCURACY,
    ATTR_LATITUDE,
    ATTR_LONGITUDE,
)
from homeassistant.components.device_tracker import SOURCE_TYPE_GPS
from homeassistant.components.device_tracker.config_entry import TrackerEntity
from homeassistant.helpers import device_registry
from homeassistant.helpers.dispatcher import async_dispatcher_connect
from homeassistant.helpers.restore_state import RestoreEntity
from homeassistant.helpers.typing import HomeAssistantType

from . import DOMAIN as GPL_DOMAIN, TRACKER_UPDATE
from .const import (
    ATTR_ACTIVITY,
    ATTR_ALTITUDE,
    ATTR_DIRECTION,
    ATTR_PROVIDER,
    ATTR_SPEED,
)

_LOGGER = logging.getLogger(__name__)


async def async_setup_entry(hass: HomeAssistantType, entry, async_add_entities):
    """Configure a dispatcher connection based on a config entry."""

    @callback
    def _receive_data(device, gps, battery, accuracy, attrs):
        """Receive set location."""
        if device in hass.data[GPL_DOMAIN]["devices"]:
            return

        hass.data[GPL_DOMAIN]["devices"].add(device)

        async_add_entities([GPSLoggerEntity(device, gps, battery, accuracy, attrs)])

    hass.data[GPL_DOMAIN]["unsub_device_tracker"][
        entry.entry_id
    ] = async_dispatcher_connect(hass, TRACKER_UPDATE, _receive_data)

    # Restore previously loaded devices
    dev_reg = await device_registry.async_get_registry(hass)
    dev_ids = {
        identifier[1]
        for device in dev_reg.devices.values()
        for identifier in device.identifiers
        if identifier[0] == GPL_DOMAIN
    }
    if not dev_ids:
        return

    entities = []
    for dev_id in dev_ids:
        hass.data[GPL_DOMAIN]["devices"].add(dev_id)
        entity = GPSLoggerEntity(dev_id, None, None, None, None)
        entities.append(entity)

    async_add_entities(entities)


class GPSLoggerEntity(TrackerEntity, RestoreEntity):
    """Represent a tracked device."""

    def __init__(self, device, location, battery, accuracy, attributes):
        """Set up Geofency entity."""
        self._accuracy = accuracy
        self._attributes = attributes
        self._name = device
        self._battery = battery
        self._location = location
        self._unsub_dispatcher = None
        self._unique_id = device

    @property
    def battery_level(self):
        """Return battery value of the device."""
        return self._battery

    @property
    def device_state_attributes(self):
        """Return device specific attributes."""
        return self._attributes

    @property
    def latitude(self):
        """Return latitude value of the device."""
        return self._location[0]

    @property
    def longitude(self):
        """Return longitude value of the device."""
        return self._location[1]

    @property
    def location_accuracy(self):
        """Return the gps accuracy of the device."""
        return self._accuracy

    @property
    def name(self):
        """Return the name of the device."""
        return self._name

    @property
    def should_poll(self):
        """No polling needed."""
        return False

    @property
    def unique_id(self):
        """Return the unique ID."""
        return self._unique_id

    @property
    def device_info(self):
        """Return the device info."""
        return {"name": self._name, "identifiers": {(GPL_DOMAIN, self._unique_id)}}

    @property
    def source_type(self):
        """Return the source type, eg gps or router, of the device."""
        return SOURCE_TYPE_GPS

    async def async_added_to_hass(self):
        """Register state update callback."""
        await super().async_added_to_hass()
        self._unsub_dispatcher = async_dispatcher_connect(
            self.hass, TRACKER_UPDATE, self._async_receive_data
        )

        # don't restore if we got created with data
        if self._location is not None:
            return

        state = await self.async_get_last_state()
        if state is None:
            self._location = (None, None)
            self._accuracy = None
            self._attributes = {
                ATTR_ALTITUDE: None,
                ATTR_ACTIVITY: None,
                ATTR_DIRECTION: None,
                ATTR_PROVIDER: None,
                ATTR_SPEED: None,
            }
            self._battery = None
            return

        attr = state.attributes
        self._location = (attr.get(ATTR_LATITUDE), attr.get(ATTR_LONGITUDE))
        self._accuracy = attr.get(ATTR_GPS_ACCURACY)
        self._attributes = {
            ATTR_ALTITUDE: attr.get(ATTR_ALTITUDE),
            ATTR_ACTIVITY: attr.get(ATTR_ACTIVITY),
            ATTR_DIRECTION: attr.get(ATTR_DIRECTION),
            ATTR_PROVIDER: attr.get(ATTR_PROVIDER),
            ATTR_SPEED: attr.get(ATTR_SPEED),
        }
        self._battery = attr.get(ATTR_BATTERY_LEVEL)

    async def async_will_remove_from_hass(self):
        """Clean up after entity before removal."""
        await super().async_will_remove_from_hass()
        self._unsub_dispatcher()

    @callback
    def _async_receive_data(self, device, location, battery, accuracy, attributes):
        """Mark the device as seen."""
        if device != self.name:
            return

        self._location = location
        self._battery = battery
        self._accuracy = accuracy
        self._attributes.update(attributes)
        self.async_write_ha_state()



"""Support for OpenTherm Gateway sensors."""
import logging

from homeassistant.components.sensor import ENTITY_ID_FORMAT
from homeassistant.core import callback
from homeassistant.helpers.dispatcher import async_dispatcher_connect
from homeassistant.helpers.entity import Entity, async_generate_entity_id

from .const import DATA_GATEWAYS, DATA_OPENTHERM_GW, SENSOR_INFO


_LOGGER = logging.getLogger(__name__)


async def async_setup_platform(hass, config, async_add_entities, discovery_info=None):
    """Set up the OpenTherm Gateway sensors."""
    if discovery_info is None:
        return
    gw_dev = hass.data[DATA_OPENTHERM_GW][DATA_GATEWAYS][discovery_info]
    sensors = []
    for var, info in SENSOR_INFO.items():
        device_class = info[0]
        unit = info[1]
        friendly_name_format = info[2]
        sensors.append(
            OpenThermSensor(gw_dev, var, device_class, unit, friendly_name_format)
        )
    async_add_entities(sensors)


class OpenThermSensor(Entity):
    """Representation of an OpenTherm Gateway sensor."""

    def __init__(self, gw_dev, var, device_class, unit, friendly_name_format):
        """Initialize the OpenTherm Gateway sensor."""
        self.entity_id = async_generate_entity_id(
            ENTITY_ID_FORMAT, "{}_{}".format(var, gw_dev.gw_id), hass=gw_dev.hass
        )
        self._gateway = gw_dev
        self._var = var
        self._value = None
        self._device_class = device_class
        self._unit = unit
        self._friendly_name = friendly_name_format.format(gw_dev.name)

    async def async_added_to_hass(self):
        """Subscribe to updates from the component."""
        _LOGGER.debug("Added OpenTherm Gateway sensor %s", self._friendly_name)
        async_dispatcher_connect(
            self.hass, self._gateway.update_signal, self.receive_report
        )

    @callback
    def receive_report(self, status):
        """Handle status updates from the component."""
        value = status.get(self._var)
        if isinstance(value, float):
            value = "{:2.1f}".format(value)
        self._value = value
        self.async_schedule_update_ha_state()

    @property
    def name(self):
        """Return the friendly name of the sensor."""
        return self._friendly_name

    @property
    def device_class(self):
        """Return the device class."""
        return self._device_class

    @property
    def state(self):
        """Return the state of the device."""
        return self._value

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement."""
        return self._unit

    @property
    def should_poll(self):
        """Return False because entity pushes its state."""
        return False



"""Offer device oriented automation."""
import voluptuous as vol

from homeassistant.const import CONF_DOMAIN, CONF_PLATFORM
from homeassistant.loader import async_get_integration


# mypy: allow-untyped-defs, no-check-untyped-defs

TRIGGER_SCHEMA = vol.Schema(
    {vol.Required(CONF_PLATFORM): "device", vol.Required(CONF_DOMAIN): str},
    extra=vol.ALLOW_EXTRA,
)


async def async_trigger(hass, config, action, automation_info):
    """Listen for trigger."""
    integration = await async_get_integration(hass, config[CONF_DOMAIN])
    platform = integration.get_platform("device_automation")
    return await platform.async_trigger(hass, config, action, automation_info)



"""Component to interface with binary sensors."""

from datetime import timedelta
import logging

import voluptuous as vol

from homeassistant.helpers.entity_component import EntityComponent
from homeassistant.helpers.entity import Entity
from homeassistant.const import STATE_ON, STATE_OFF
from homeassistant.helpers.config_validation import (  # noqa
    PLATFORM_SCHEMA,
    PLATFORM_SCHEMA_BASE,
)


# mypy: allow-untyped-defs, no-check-untyped-defs

DOMAIN = "binary_sensor"
SCAN_INTERVAL = timedelta(seconds=30)

ENTITY_ID_FORMAT = DOMAIN + ".{}"

# On means low, Off means normal
DEVICE_CLASS_BATTERY = "battery"

# On means cold, Off means normal
DEVICE_CLASS_COLD = "cold"

# On means connected, Off means disconnected
DEVICE_CLASS_CONNECTIVITY = "connectivity"

# On means open, Off means closed
DEVICE_CLASS_DOOR = "door"

# On means open, Off means closed
DEVICE_CLASS_GARAGE_DOOR = "garage_door"

# On means gas detected, Off means no gas (clear)
DEVICE_CLASS_GAS = "gas"

# On means hot, Off means normal
DEVICE_CLASS_HEAT = "heat"

# On means light detected, Off means no light
DEVICE_CLASS_LIGHT = "light"

# On means open (unlocked), Off means closed (locked)
DEVICE_CLASS_LOCK = "lock"

# On means wet, Off means dry
DEVICE_CLASS_MOISTURE = "moisture"

# On means motion detected, Off means no motion (clear)
DEVICE_CLASS_MOTION = "motion"

# On means moving, Off means not moving (stopped)
DEVICE_CLASS_MOVING = "moving"

# On means occupied, Off means not occupied (clear)
DEVICE_CLASS_OCCUPANCY = "occupancy"

# On means open, Off means closed
DEVICE_CLASS_OPENING = "opening"

# On means plugged in, Off means unplugged
DEVICE_CLASS_PLUG = "plug"

# On means power detected, Off means no power
DEVICE_CLASS_POWER = "power"

# On means home, Off means away
DEVICE_CLASS_PRESENCE = "presence"

# On means problem detected, Off means no problem (OK)
DEVICE_CLASS_PROBLEM = "problem"

# On means unsafe, Off means safe
DEVICE_CLASS_SAFETY = "safety"

# On means smoke detected, Off means no smoke (clear)
DEVICE_CLASS_SMOKE = "smoke"

# On means sound detected, Off means no sound (clear)
DEVICE_CLASS_SOUND = "sound"

# On means vibration detected, Off means no vibration
DEVICE_CLASS_VIBRATION = "vibration"

# On means open, Off means closed
DEVICE_CLASS_WINDOW = "window"

DEVICE_CLASSES = [
    DEVICE_CLASS_BATTERY,
    DEVICE_CLASS_COLD,
    DEVICE_CLASS_CONNECTIVITY,
    DEVICE_CLASS_DOOR,
    DEVICE_CLASS_GARAGE_DOOR,
    DEVICE_CLASS_GAS,
    DEVICE_CLASS_HEAT,
    DEVICE_CLASS_LIGHT,
    DEVICE_CLASS_LOCK,
    DEVICE_CLASS_MOISTURE,
    DEVICE_CLASS_MOTION,
    DEVICE_CLASS_MOVING,
    DEVICE_CLASS_OCCUPANCY,
    DEVICE_CLASS_OPENING,
    DEVICE_CLASS_PLUG,
    DEVICE_CLASS_POWER,
    DEVICE_CLASS_PRESENCE,
    DEVICE_CLASS_PROBLEM,
    DEVICE_CLASS_SAFETY,
    DEVICE_CLASS_SMOKE,
    DEVICE_CLASS_SOUND,
    DEVICE_CLASS_VIBRATION,
    DEVICE_CLASS_WINDOW,
]

DEVICE_CLASSES_SCHEMA = vol.All(vol.Lower, vol.In(DEVICE_CLASSES))


async def async_setup(hass, config):
    """Track states and offer events for binary sensors."""
    component = hass.data[DOMAIN] = EntityComponent(
        logging.getLogger(__name__), DOMAIN, hass, SCAN_INTERVAL
    )

    await component.async_setup(config)
    return True


async def async_setup_entry(hass, entry):
    """Set up a config entry."""
    return await hass.data[DOMAIN].async_setup_entry(entry)


async def async_unload_entry(hass, entry):
    """Unload a config entry."""
    return await hass.data[DOMAIN].async_unload_entry(entry)


class BinarySensorDevice(Entity):
    """Represent a binary sensor."""

    @property
    def is_on(self):
        """Return true if the binary sensor is on."""
        return None

    @property
    def state(self):
        """Return the state of the binary sensor."""
        return STATE_ON if self.is_on else STATE_OFF

    @property
    def device_class(self):
        """Return the class of this device, from component DEVICE_CLASSES."""
        return None



"""Support for the Hive switches."""
from homeassistant.components.switch import SwitchDevice

from . import DATA_HIVE, DOMAIN


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up Hive switches."""
    if discovery_info is None:
        return
    session = hass.data.get(DATA_HIVE)

    add_entities([HiveDevicePlug(session, discovery_info)])


class HiveDevicePlug(SwitchDevice):
    """Hive Active Plug."""

    def __init__(self, hivesession, hivedevice):
        """Initialize the Switch device."""
        self.node_id = hivedevice["Hive_NodeID"]
        self.node_name = hivedevice["Hive_NodeName"]
        self.device_type = hivedevice["HA_DeviceType"]
        self.session = hivesession
        self.attributes = {}
        self.data_updatesource = "{}.{}".format(self.device_type, self.node_id)
        self._unique_id = "{}-{}".format(self.node_id, self.device_type)
        self.session.entities.append(self)

    @property
    def unique_id(self):
        """Return unique ID of entity."""
        return self._unique_id

    @property
    def device_info(self):
        """Return device information."""
        return {"identifiers": {(DOMAIN, self.unique_id)}, "name": self.name}

    def handle_update(self, updatesource):
        """Handle the new update request."""
        if "{}.{}".format(self.device_type, self.node_id) not in updatesource:
            self.schedule_update_ha_state()

    @property
    def name(self):
        """Return the name of this Switch device if any."""
        return self.node_name

    @property
    def device_state_attributes(self):
        """Show Device Attributes."""
        return self.attributes

    @property
    def current_power_w(self):
        """Return the current power usage in W."""
        return self.session.switch.get_power_usage(self.node_id)

    @property
    def is_on(self):
        """Return true if switch is on."""
        return self.session.switch.get_state(self.node_id)

    def turn_on(self, **kwargs):
        """Turn the switch on."""
        self.session.switch.turn_on(self.node_id)
        for entity in self.session.entities:
            entity.handle_update(self.data_updatesource)

    def turn_off(self, **kwargs):
        """Turn the device off."""
        self.session.switch.turn_off(self.node_id)
        for entity in self.session.entities:
            entity.handle_update(self.data_updatesource)

    def update(self):
        """Update all Node data from Hive."""
        self.session.core.update_data(self.node_id)
        self.attributes = self.session.attributes.state_attributes(self.node_id)



"""Sensor for UPS packages."""
import logging
from collections import defaultdict
from datetime import timedelta

import voluptuous as vol

import homeassistant.helpers.config_validation as cv
from homeassistant.components.sensor import PLATFORM_SCHEMA
from homeassistant.const import (
    ATTR_ATTRIBUTION,
    CONF_NAME,
    CONF_PASSWORD,
    CONF_SCAN_INTERVAL,
    CONF_USERNAME,
)
from homeassistant.helpers.entity import Entity
from homeassistant.util import Throttle, slugify
from homeassistant.util.dt import now, parse_date

_LOGGER = logging.getLogger(__name__)

DOMAIN = "ups"
COOKIE = "upsmychoice_cookies.pickle"
ICON = "mdi:package-variant-closed"
STATUS_DELIVERED = "delivered"

SCAN_INTERVAL = timedelta(seconds=1800)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_USERNAME): cv.string,
        vol.Required(CONF_PASSWORD): cv.string,
        vol.Optional(CONF_NAME): cv.string,
    }
)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the UPS platform."""
    import upsmychoice

    _LOGGER.warning(
        "The ups integration is deprecated and will be removed "
        "in Home Assistant 0.100.0. For more information see ADR-0004:"
        "https://github.com/home-assistant/architecture/blob/master/adr/0004-webscraping.md"
    )

    try:
        cookie = hass.config.path(COOKIE)
        session = upsmychoice.get_session(
            config.get(CONF_USERNAME), config.get(CONF_PASSWORD), cookie_path=cookie
        )
    except upsmychoice.UPSError:
        _LOGGER.exception("Could not connect to UPS My Choice")
        return False

    add_entities(
        [
            UPSSensor(
                session,
                config.get(CONF_NAME),
                config.get(CONF_SCAN_INTERVAL, SCAN_INTERVAL),
            )
        ],
        True,
    )


class UPSSensor(Entity):
    """UPS Sensor."""

    def __init__(self, session, name, interval):
        """Initialize the sensor."""
        self._session = session
        self._name = name
        self._attributes = None
        self._state = None
        self.update = Throttle(interval)(self._update)

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name or DOMAIN

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement of this entity, if any."""
        return "packages"

    def _update(self):
        """Update device state."""
        import upsmychoice

        status_counts = defaultdict(int)
        try:
            for package in upsmychoice.get_packages(self._session):
                status = slugify(package["status"])
                skip = (
                    status == STATUS_DELIVERED
                    and parse_date(package["delivery_date"]) < now().date()
                )
                if skip:
                    continue
                status_counts[status] += 1
        except upsmychoice.UPSError:
            _LOGGER.error("Could not connect to UPS My Choice account")

        self._attributes = {ATTR_ATTRIBUTION: upsmychoice.ATTRIBUTION}
        self._attributes.update(status_counts)
        self._state = sum(status_counts.values())

    @property
    def device_state_attributes(self):
        """Return the state attributes."""
        return self._attributes

    @property
    def icon(self):
        """Icon to use in the frontend."""
        return ICON



"""Provide functionality to TTS."""
import asyncio
import ctypes
import functools as ft
import hashlib
import io
import logging
import mimetypes
import os
import re
from typing import Optional

from aiohttp import web
import voluptuous as vol

from homeassistant.components.http import HomeAssistantView
from homeassistant.components.media_player.const import (
    ATTR_MEDIA_CONTENT_ID,
    ATTR_MEDIA_CONTENT_TYPE,
    MEDIA_TYPE_MUSIC,
    SERVICE_PLAY_MEDIA,
)
from homeassistant.components.media_player.const import DOMAIN as DOMAIN_MP
from homeassistant.const import ATTR_ENTITY_ID, ENTITY_MATCH_ALL, CONF_PLATFORM
from homeassistant.core import callback
from homeassistant.exceptions import HomeAssistantError
from homeassistant.helpers import config_per_platform
import homeassistant.helpers.config_validation as cv
from homeassistant.helpers.typing import HomeAssistantType
from homeassistant.setup import async_prepare_setup_platform


# mypy: allow-untyped-defs, no-check-untyped-defs

_LOGGER = logging.getLogger(__name__)

ATTR_CACHE = "cache"
ATTR_LANGUAGE = "language"
ATTR_MESSAGE = "message"
ATTR_OPTIONS = "options"
ATTR_PLATFORM = "platform"

CONF_BASE_URL = "base_url"
CONF_CACHE = "cache"
CONF_CACHE_DIR = "cache_dir"
CONF_LANG = "language"
CONF_SERVICE_NAME = "service_name"
CONF_TIME_MEMORY = "time_memory"

DEFAULT_CACHE = True
DEFAULT_CACHE_DIR = "tts"
DEFAULT_TIME_MEMORY = 300
DOMAIN = "tts"

MEM_CACHE_FILENAME = "filename"
MEM_CACHE_VOICE = "voice"

SERVICE_CLEAR_CACHE = "clear_cache"
SERVICE_SAY = "say"

_RE_VOICE_FILE = re.compile(r"([a-f0-9]{40})_([^_]+)_([^_]+)_([a-z_]+)\.[a-z0-9]{3,4}")
KEY_PATTERN = "{0}_{1}_{2}_{3}"


def _deprecated_platform(value):
    """Validate if platform is deprecated."""
    if value == "google":
        raise vol.Invalid(
            "google tts service has been renamed to google_translate,"
            " please update your configuration."
        )
    return value


PLATFORM_SCHEMA = cv.PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_PLATFORM): vol.All(cv.string, _deprecated_platform),
        vol.Optional(CONF_CACHE, default=DEFAULT_CACHE): cv.boolean,
        vol.Optional(CONF_CACHE_DIR, default=DEFAULT_CACHE_DIR): cv.string,
        vol.Optional(CONF_TIME_MEMORY, default=DEFAULT_TIME_MEMORY): vol.All(
            vol.Coerce(int), vol.Range(min=60, max=57600)
        ),
        vol.Optional(CONF_BASE_URL): cv.string,
        vol.Optional(CONF_SERVICE_NAME): cv.string,
    }
)
PLATFORM_SCHEMA_BASE = cv.PLATFORM_SCHEMA_BASE.extend(PLATFORM_SCHEMA.schema)

SCHEMA_SERVICE_SAY = vol.Schema(
    {
        vol.Required(ATTR_MESSAGE): cv.string,
        vol.Optional(ATTR_CACHE): cv.boolean,
        vol.Optional(ATTR_ENTITY_ID): cv.comp_entity_ids,
        vol.Optional(ATTR_LANGUAGE): cv.string,
        vol.Optional(ATTR_OPTIONS): dict,
    }
)

SCHEMA_SERVICE_CLEAR_CACHE = vol.Schema({})


async def async_setup(hass, config):
    """Set up TTS."""
    tts = SpeechManager(hass)

    try:
        conf = config[DOMAIN][0] if config.get(DOMAIN, []) else {}
        use_cache = conf.get(CONF_CACHE, DEFAULT_CACHE)
        cache_dir = conf.get(CONF_CACHE_DIR, DEFAULT_CACHE_DIR)
        time_memory = conf.get(CONF_TIME_MEMORY, DEFAULT_TIME_MEMORY)
        base_url = conf.get(CONF_BASE_URL) or hass.config.api.base_url

        await tts.async_init_cache(use_cache, cache_dir, time_memory, base_url)
    except (HomeAssistantError, KeyError) as err:
        _LOGGER.error("Error on cache init %s", err)
        return False

    hass.http.register_view(TextToSpeechView(tts))
    hass.http.register_view(TextToSpeechUrlView(tts))

    async def async_setup_platform(p_type, p_config, disc_info=None):
        """Set up a TTS platform."""
        platform = await async_prepare_setup_platform(hass, config, DOMAIN, p_type)
        if platform is None:
            return

        try:
            if hasattr(platform, "async_get_engine"):
                provider = await platform.async_get_engine(hass, p_config)
            else:
                provider = await hass.async_add_job(platform.get_engine, hass, p_config)

            if provider is None:
                _LOGGER.error("Error setting up platform %s", p_type)
                return

            tts.async_register_engine(p_type, provider, p_config)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Error setting up platform: %s", p_type)
            return

        async def async_say_handle(service):
            """Service handle for say."""
            entity_ids = service.data.get(ATTR_ENTITY_ID, ENTITY_MATCH_ALL)
            message = service.data.get(ATTR_MESSAGE)
            cache = service.data.get(ATTR_CACHE)
            language = service.data.get(ATTR_LANGUAGE)
            options = service.data.get(ATTR_OPTIONS)

            try:
                url = await tts.async_get_url(
                    p_type, message, cache=cache, language=language, options=options
                )
            except HomeAssistantError as err:
                _LOGGER.error("Error on init TTS: %s", err)
                return

            data = {
                ATTR_MEDIA_CONTENT_ID: url,
                ATTR_MEDIA_CONTENT_TYPE: MEDIA_TYPE_MUSIC,
                ATTR_ENTITY_ID: entity_ids,
            }

            await hass.services.async_call(
                DOMAIN_MP, SERVICE_PLAY_MEDIA, data, blocking=True
            )

        service_name = p_config.get(
            CONF_SERVICE_NAME, "{}_{}".format(p_type, SERVICE_SAY)
        )
        hass.services.async_register(
            DOMAIN, service_name, async_say_handle, schema=SCHEMA_SERVICE_SAY
        )

    setup_tasks = [
        async_setup_platform(p_type, p_config)
        for p_type, p_config in config_per_platform(config, DOMAIN)
    ]

    if setup_tasks:
        await asyncio.wait(setup_tasks)

    async def async_clear_cache_handle(service):
        """Handle clear cache service call."""
        await tts.async_clear_cache()

    hass.services.async_register(
        DOMAIN,
        SERVICE_CLEAR_CACHE,
        async_clear_cache_handle,
        schema=SCHEMA_SERVICE_CLEAR_CACHE,
    )

    return True


class SpeechManager:
    """Representation of a speech store."""

    def __init__(self, hass):
        """Initialize a speech store."""
        self.hass = hass
        self.providers = {}

        self.use_cache = DEFAULT_CACHE
        self.cache_dir = DEFAULT_CACHE_DIR
        self.time_memory = DEFAULT_TIME_MEMORY
        self.base_url = None
        self.file_cache = {}
        self.mem_cache = {}

    async def async_init_cache(self, use_cache, cache_dir, time_memory, base_url):
        """Init config folder and load file cache."""
        self.use_cache = use_cache
        self.time_memory = time_memory
        self.base_url = base_url

        def init_tts_cache_dir(cache_dir):
            """Init cache folder."""
            if not os.path.isabs(cache_dir):
                cache_dir = self.hass.config.path(cache_dir)
            if not os.path.isdir(cache_dir):
                _LOGGER.info("Create cache dir %s.", cache_dir)
                os.mkdir(cache_dir)
            return cache_dir

        try:
            self.cache_dir = await self.hass.async_add_job(
                init_tts_cache_dir, cache_dir
            )
        except OSError as err:
            raise HomeAssistantError("Can't init cache dir {}".format(err))

        def get_cache_files():
            """Return a dict of given engine files."""
            cache = {}

            folder_data = os.listdir(self.cache_dir)
            for file_data in folder_data:
                record = _RE_VOICE_FILE.match(file_data)
                if record:
                    key = KEY_PATTERN.format(
                        record.group(1),
                        record.group(2),
                        record.group(3),
                        record.group(4),
                    )
                    cache[key.lower()] = file_data.lower()
            return cache

        try:
            cache_files = await self.hass.async_add_job(get_cache_files)
        except OSError as err:
            raise HomeAssistantError("Can't read cache dir {}".format(err))

        if cache_files:
            self.file_cache.update(cache_files)

    async def async_clear_cache(self):
        """Read file cache and delete files."""
        self.mem_cache = {}

        def remove_files():
            """Remove files from filesystem."""
            for _, filename in self.file_cache.items():
                try:
                    os.remove(os.path.join(self.cache_dir, filename))
                except OSError as err:
                    _LOGGER.warning("Can't remove cache file '%s': %s", filename, err)

        await self.hass.async_add_job(remove_files)
        self.file_cache = {}

    @callback
    def async_register_engine(self, engine, provider, config):
        """Register a TTS provider."""
        provider.hass = self.hass
        if provider.name is None:
            provider.name = engine
        self.providers[engine] = provider

    async def async_get_url(
        self, engine, message, cache=None, language=None, options=None
    ):
        """Get URL for play message.

        This method is a coroutine.
        """
        provider = self.providers[engine]
        msg_hash = hashlib.sha1(bytes(message, "utf-8")).hexdigest()
        use_cache = cache if cache is not None else self.use_cache

        # Languages
        language = language or provider.default_language
        if language is None or language not in provider.supported_languages:
            raise HomeAssistantError("Not supported language {0}".format(language))

        # Options
        if provider.default_options and options:
            merged_options = provider.default_options.copy()
            merged_options.update(options)
            options = merged_options
        options = options or provider.default_options
        if options is not None:
            invalid_opts = [
                opt_name
                for opt_name in options.keys()
                if opt_name not in (provider.supported_options or [])
            ]
            if invalid_opts:
                raise HomeAssistantError(
                    "Invalid options found: {}".format(invalid_opts)
                )
            options_key = ctypes.c_size_t(hash(frozenset(options))).value
        else:
            options_key = "-"

        key = KEY_PATTERN.format(msg_hash, language, options_key, engine).lower()

        # Is speech already in memory
        if key in self.mem_cache:
            filename = self.mem_cache[key][MEM_CACHE_FILENAME]
        # Is file store in file cache
        elif use_cache and key in self.file_cache:
            filename = self.file_cache[key]
            self.hass.async_create_task(self.async_file_to_mem(key))
        # Load speech from provider into memory
        else:
            filename = await self.async_get_tts_audio(
                engine, key, message, use_cache, language, options
            )

        return "{}/api/tts_proxy/{}".format(self.base_url, filename)

    async def async_get_tts_audio(self, engine, key, message, cache, language, options):
        """Receive TTS and store for view in cache.

        This method is a coroutine.
        """
        provider = self.providers[engine]
        extension, data = await provider.async_get_tts_audio(message, language, options)

        if data is None or extension is None:
            raise HomeAssistantError("No TTS from {} for '{}'".format(engine, message))

        # Create file infos
        filename = ("{}.{}".format(key, extension)).lower()

        data = self.write_tags(filename, data, provider, message, language, options)

        # Save to memory
        self._async_store_to_memcache(key, filename, data)

        if cache:
            self.hass.async_create_task(self.async_save_tts_audio(key, filename, data))

        return filename

    async def async_save_tts_audio(self, key, filename, data):
        """Store voice data to file and file_cache.

        This method is a coroutine.
        """
        voice_file = os.path.join(self.cache_dir, filename)

        def save_speech():
            """Store speech to filesystem."""
            with open(voice_file, "wb") as speech:
                speech.write(data)

        try:
            await self.hass.async_add_job(save_speech)
            self.file_cache[key] = filename
        except OSError:
            _LOGGER.error("Can't write %s", filename)

    async def async_file_to_mem(self, key):
        """Load voice from file cache into memory.

        This method is a coroutine.
        """
        filename = self.file_cache.get(key)
        if not filename:
            raise HomeAssistantError("Key {} not in file cache!".format(key))

        voice_file = os.path.join(self.cache_dir, filename)

        def load_speech():
            """Load a speech from filesystem."""
            with open(voice_file, "rb") as speech:
                return speech.read()

        try:
            data = await self.hass.async_add_job(load_speech)
        except OSError:
            del self.file_cache[key]
            raise HomeAssistantError("Can't read {}".format(voice_file))

        self._async_store_to_memcache(key, filename, data)

    @callback
    def _async_store_to_memcache(self, key, filename, data):
        """Store data to memcache and set timer to remove it."""
        self.mem_cache[key] = {MEM_CACHE_FILENAME: filename, MEM_CACHE_VOICE: data}

        @callback
        def async_remove_from_mem():
            """Cleanup memcache."""
            self.mem_cache.pop(key)

        self.hass.loop.call_later(self.time_memory, async_remove_from_mem)

    async def async_read_tts(self, filename):
        """Read a voice file and return binary.

        This method is a coroutine.
        """
        record = _RE_VOICE_FILE.match(filename.lower())
        if not record:
            raise HomeAssistantError("Wrong tts file format!")

        key = KEY_PATTERN.format(
            record.group(1), record.group(2), record.group(3), record.group(4)
        )

        if key not in self.mem_cache:
            if key not in self.file_cache:
                raise HomeAssistantError("{} not in cache!".format(key))
            await self.async_file_to_mem(key)

        content, _ = mimetypes.guess_type(filename)
        return (content, self.mem_cache[key][MEM_CACHE_VOICE])

    @staticmethod
    def write_tags(filename, data, provider, message, language, options):
        """Write ID3 tags to file.

        Async friendly.
        """
        import mutagen

        data_bytes = io.BytesIO(data)
        data_bytes.name = filename
        data_bytes.seek(0)

        album = provider.name
        artist = language

        if options is not None:
            if options.get("voice") is not None:
                artist = options.get("voice")

        try:
            tts_file = mutagen.File(data_bytes, easy=True)
            if tts_file is not None:
                tts_file["artist"] = artist
                tts_file["album"] = album
                tts_file["title"] = message
                tts_file.save(data_bytes)
        except mutagen.MutagenError as err:
            _LOGGER.error("ID3 tag error: %s", err)

        return data_bytes.getvalue()


class Provider:
    """Represent a single TTS provider."""

    hass: Optional[HomeAssistantType] = None
    name: Optional[str] = None

    @property
    def default_language(self):
        """Return the default language."""
        return None

    @property
    def supported_languages(self):
        """Return a list of supported languages."""
        return None

    @property
    def supported_options(self):
        """Return a list of supported options like voice, emotionen."""
        return None

    @property
    def default_options(self):
        """Return a dict include default options."""
        return None

    def get_tts_audio(self, message, language, options=None):
        """Load tts audio file from provider."""
        raise NotImplementedError()

    def async_get_tts_audio(self, message, language, options=None):
        """Load tts audio file from provider.

        Return a tuple of file extension and data as bytes.

        This method must be run in the event loop and returns a coroutine.
        """
        return self.hass.async_add_job(
            ft.partial(self.get_tts_audio, message, language, options=options)
        )


class TextToSpeechUrlView(HomeAssistantView):
    """TTS view to get a url to a generated speech file."""

    requires_auth = True
    url = "/api/tts_get_url"
    name = "api:tts:geturl"

    def __init__(self, tts):
        """Initialize a tts view."""
        self.tts = tts

    async def post(self, request):
        """Generate speech and provide url."""
        try:
            data = await request.json()
        except ValueError:
            return self.json_message("Invalid JSON specified", 400)
        if not data.get(ATTR_PLATFORM) and data.get(ATTR_MESSAGE):
            return self.json_message("Must specify platform and message", 400)

        p_type = data[ATTR_PLATFORM]
        message = data[ATTR_MESSAGE]
        cache = data.get(ATTR_CACHE)
        language = data.get(ATTR_LANGUAGE)
        options = data.get(ATTR_OPTIONS)

        try:
            url = await self.tts.async_get_url(
                p_type, message, cache=cache, language=language, options=options
            )
            resp = self.json({"url": url}, 200)
        except HomeAssistantError as err:
            _LOGGER.error("Error on init tts: %s", err)
            resp = self.json({"error": err}, 400)

        return resp


class TextToSpeechView(HomeAssistantView):
    """TTS view to serve a speech audio."""

    requires_auth = False
    url = "/api/tts_proxy/{filename}"
    name = "api:tts:speech"

    def __init__(self, tts):
        """Initialize a tts view."""
        self.tts = tts

    async def get(self, request, filename):
        """Start a get request."""
        try:
            content, data = await self.tts.async_read_tts(filename)
        except HomeAssistantError as err:
            _LOGGER.error("Error on load tts: %s", err)
            return web.Response(status=404)

        return web.Response(body=data, content_type=content)
